﻿using System;
using System.IO;
using System.Xml;
using System.Xml.Schema;
using System.Threading;

using ElementaryExtensions;

namespace ShaperTerminal {

//**************************************************************************************************
class XmlConfigurationException: ЄException {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

public XmlConfigurationException(string strMessage)
   :base(strMessage)
{}

//--------------------------------------------------------------------------------------------------

public static new XmlConfigurationException Create(string strMessageFmt,params object[] aobParams)
{
   return new XmlConfigurationException(string.Format(strMessageFmt,aobParams));
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XmlConfigurationException
//**************************************************************************************************


//**************************************************************************************************
abstract class XmlConfiguration {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

protected XmlConfiguration(string strSchemaText)
{
   using (StringReader stringReader=new StringReader(strSchemaText))
      i_Schema=XmlSchema.Read(stringReader,null);
}

//--------------------------------------------------------------------------------------------------

protected XmlDocument Document
{
   get { return i_Document; }
}

//--------------------------------------------------------------------------------------------------

protected XmlNode CurrentNode = null;

//==================================================================================================

protected void LoadDocument(string strFileName)
{
   XmlSchemaSet
      schemaSet=new XmlSchemaSet();
   XmlReaderSettings
      readerSettings=new XmlReaderSettings();

   schemaSet.Add(i_Schema);

   readerSettings.ValidationType=ValidationType.Schema;
   readerSettings.Schemas=schemaSet;

   i_strDocumentFilePath = strFileName;

   try
   {
      Interlocked.Exchange(
         ref i_strDocumentFilePath,
         Path.GetFullPath(strFileName) );

      using (XmlReader reader=XmlReader.Create(i_strDocumentFilePath,readerSettings))
         i_Document.Load(reader);

      if ( i_Document.DocumentElement.NamespaceURI != i_Schema.TargetNamespace )
         throw new ApplicationException(string.Format(
            "Элемент документа должен относиться к следующему пространству имен: \"{0}\".",
            i_Schema.TargetNamespace ));
   }
   catch (FileNotFoundException e)
   {
      throw new XmlConfigurationException(e.Message);
   }
   catch (Exception e)
   {
      throw XmlConfigurationException.Create(
         "{0}\n"+
         "Файл: \"{1}\".",
         e.Message, i_strDocumentFilePath );
   }

   i_NamespaceManager=new XmlNamespaceManager(i_Document.NameTable);
   i_NamespaceManager.AddNamespace("x",i_Schema.TargetNamespace);

   CurrentNode=i_Document.DocumentElement;
}

//--------------------------------------------------------------------------------------------------

protected bool NavigateToNode(string strDocElementRelativePath)
{
   return NavigateToNode(strDocElementRelativePath,false);
}

protected bool NavigateToNode(string strDocElementRelativePath,bool bFailIfNotFound)
{
   string
      strNodePath = string.Format( "/x:{0}/{1}",
                                   i_Document.DocumentElement.Name,
                                   strDocElementRelativePath );

   CurrentNode = i_Document.SelectSingleNode(strNodePath,i_NamespaceManager);

   if (CurrentNode!=null)
      i_strCurrentNodePath=strNodePath;
   else
   {
      i_strCurrentNodePath=null;
      if (bFailIfNotFound)
         throw CreateExceptionWithFile(
            "Узел не найден: \"/x:{0}/{1}\".",
            i_Document.DocumentElement.Name,
            strDocElementRelativePath );
   }

   return CurrentNode!=null;
}

protected void VerifyCurrentNode()
{
   if (CurrentNode==null)
      throw new InvalidOperationException("Текущий узел XML не установлен.");
}

//==================================================================================================

protected XmlConfigurationException CreateExceptionWithPath(string strMessage)
{
   VerifyCurrentNode();

   return
      XmlConfigurationException.Create(
         "{0}\n"+
         "Файл: \"{1}\".\n"+
         "Путь узла: \"{2}\".",
         strMessage, i_strDocumentFilePath, i_strCurrentNodePath );
}

protected XmlConfigurationException CreateExceptionWithPath( string strMessageFmt,
                                                             params object[] aobParams )
{
   return CreateExceptionWithPath(string.Format(strMessageFmt,aobParams));
}

//--------------------------------------------------------------------------------------------------

protected XmlConfigurationException CreateExceptionWithPath( XmlAttribute attribute,
                                                             string strMessage )
{
   VerifyCurrentNode();

   return
      XmlConfigurationException.Create(
         "{0}\n"+
         "Файл: \"{1}\".\n"+
         "Путь узла: \"{2}\".\n"+
         "Атрибут: \"{3}\".",
         strMessage, i_strDocumentFilePath, i_strCurrentNodePath , attribute.Name );
}

protected XmlConfigurationException CreateExceptionWithPath( XmlAttribute attribute,
                                                             string strMessageFmt,
                                                             params object[] aobParams )
{
   return CreateExceptionWithPath(attribute,string.Format(strMessageFmt,aobParams));
}

//==================================================================================================

protected XmlConfigurationException CreateExceptionWithFile(string strMessage)
{
   return
      XmlConfigurationException.Create(
         "{0}\n"+
         "Файл: \"{1}\".",
         strMessage, i_strDocumentFilePath, i_strCurrentNodePath );
}

protected XmlConfigurationException CreateExceptionWithFile( string strMessageFmt,
                                                             params object[] aobParams )
{
   return CreateExceptionWithFile(string.Format(strMessageFmt,aobParams));
}

//==================================================================================================

static protected XmlNode[] GetChildElements(XmlNode node)
{
   XmlNodeList
      nodeList=node.ChildNodes;
   int
      nElementCount=0;
   foreach (XmlNode childNode in nodeList)
      if (childNode.NodeType==XmlNodeType.Element)
         nElementCount++;
   XmlNode[]
      elements=new XmlNode[nElementCount];
   for (int i=0,j=0; i<nodeList.Count; i++)
      if (nodeList[i].NodeType==XmlNodeType.Element)
         elements[j++]=nodeList[i];
   return elements;
}

//==================================================================================================

XmlSchema
   i_Schema;
XmlDocument
   i_Document=new XmlDocument();
XmlNamespaceManager
   i_NamespaceManager;
string
   i_strDocumentFilePath,
   i_strCurrentNodePath;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XmlConfiguration
//**************************************************************************************************

} // XmlConfiguration
