﻿using System;
using System.Threading;
using System.Reflection;
using System.Globalization;

using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Threading;
using System.Windows.Data;

using XHelpers;

namespace PaycorMaster {

using AppProperties = Properties;

//**************************************************************************************************
public partial class App: Application {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

static AppSettings
   _Settings_;
static AppConfiguration
   _Configuration_;
static DbConnector
   _DbConnector_;
static string
   _ConfigurationFileName_;
static string[]
   _StartupArguments_;
static CustomPage
   _SelectedPage_;
static bool
   _IsInitialized_,
   _HasNonSavedSettings_;

//==================================================================================================

public const int
   FailureExitCode = 1;

//--------------------------------------------------------------------------------------------------

public static string Title
   { get { return AssemblyInfo.Title; } }

public new static App Current
   { get { return (App)Application.Current; } }

public new static MainWindow MainWindow
   { get { return (MainWindow)Application.Current.MainWindow; } }

public static bool IsPageSelected
   { get { return _SelectedPage_!=null; } }

public static CustomPage SelectedPage
{
   get
   {
      if (!IsPageSelected)
         throw new ApplicationException("Страница навигации не выбрана.");
      return _SelectedPage_;
   }
}

//--------------------------------------------------------------------------------------------------

public static AppSettings Settings
   { get { return _Settings_; } }

public static bool HasNonsavedSettings
   { get { return _HasNonSavedSettings_; } }

public static AppConfiguration Configuration
   { get { return _Configuration_; } }

public static string ConfigurationFileName
   { get { return _ConfigurationFileName_; } }

//--------------------------------------------------------------------------------------------------

public static bool IsInitialized
   { get { return _IsInitialized_; } }

public static bool IsConnectedToDatabase
   { get { return _DbConnector_.IsConnected; } }

public static PaycorMasterDataContext Database
   { get { return _DbConnector_.Database; } }

public static DbConnector DbConnector
   { get { return _DbConnector_; } }

public static bool IsInAdminMode
   { get { return _DbConnector_.IsUserInSysadminRole; } }

//==================================================================================================

public static void Navigate(CustomPage page)
{
   PrepareToNavigation(); DoNavigate(page);
}

public static void Navigate(CustomPage page,string title)
{
   PrepareToNavigation(); DoNavigate(page,title);
}

//--------------------------------------------------------------------------------------------------

public static void Navigate(Type pageType)
{
   VerifyPageType( pageType );
   PrepareToNavigation();
   XDispatcher.BeginInvoke(delegate{
      DoNavigate( (CustomPage)XType.CreateInstance(pageType) );
   });
}

public static void Navigate(Type pageType,string title)
{
   VerifyPageType( pageType );
   PrepareToNavigation();
   XDispatcher.BeginInvoke(delegate{
      DoNavigate( (CustomPage)XType.CreateInstance(pageType), title );
   });
}

//==================================================================================================

public static bool Initialize()
{
   using (new XAppWaiting())
   try
   {
      XAppWaiting.DisplayStatus(
         "Проверка настроек приложения (пользовательский профиль)" );
      if (_Settings_.InitializationException!=null)
         throw new XOperationFailedException(
            XException.ComposeChainedMessage(
               _Settings_.InitializationException ) );
      XAppWaiting.DisplayStatus(
         "Разбор параметров командной строки" );
      switch (_StartupArguments_.Length)
      {
      case 0:
         _ConfigurationFileName_ = XProgram.SimpleName+".xml";
         break;
      case 1:
         _ConfigurationFileName_ = _StartupArguments_[0];
         break;
      default:
         throw new XOperationFailedException(
            "Некорректные параметры коммандной строки.\r\n"+
            "Допустим лишь один (необязательный) параметр \u2014 имя файла конфигурации." );
      }

      XAppWaiting.DisplayStatus("Извлечение XSD из ресурсов");
      _Configuration_ = new AppConfiguration();

      XAppWaiting.DisplayStatus(
         "Чтение конфигурации: \"{0}\"",
         _ConfigurationFileName_ );
      Configuration.Load( _ConfigurationFileName_ );

      _DbConnector_.ServerName        = _Configuration_.ServerName;
      _DbConnector_.DatabaseName      = _Configuration_.DatabaseName;
      _DbConnector_.IntegratedSecutity= _Configuration_.SystemAuthentication;
      _DbConnector_.ApplicationName   = XProgram.SimpleName;
   }
   catch (Exception e)
   {
      XAppWaiting.ReportError(e,false);
      XAppWaiting.Activate( AppCommands.Close.Text );
      return false;
   }
   _StartupArguments_ = null;
   return _IsInitialized_ = true;
}

//--------------------------------------------------------------------------------------------------

public static void ApplySettings(AppSettings settings)
{
   App._Settings_ = settings;
   Current.SynchronizeSettings();
   _HasNonSavedSettings_ = true;
}

public static bool ApplySettings(AppSettings settings,bool save)
{
   if (!save)
   {
      ApplySettings(settings);
      return true;
   }
   try
   {
      settings.Save();
   }
   catch (Exception e)
   {
      XAppWaiting.ReportError(e);
      return false;
   }
   ApplySettings(settings);
   _HasNonSavedSettings_ = false;
   return true;
}

//--------------------------------------------------------------------------------------------------

public static bool ConnectDatabase()
{
   var resultCapture = new XSyncCapture<bool>();
   XDispatcher.Synchronize(delegate{
      resultCapture.Value = LoginWindow.ShowDialog();
   });
   return resultCapture;
}

public static bool DisconnectDatabase()
{
   if (!IsConnectedToDatabase)
      return true;
   using (new XAppWaiting("Отключение от базы данных"))
   {
      Exception e;
      bool ok = _DbConnector_.Disconnect(out e);
      if (!ok) XAppWaiting.ReportError( e );
      XDispatcher.Synchronize(delegate{
         MainWindow.SetConnectionInformation(); });
      return ok;
   }
}

//--------------------------------------------------------------------------------------------------

public static bool EstablishDbConnection()
{
   return EstablishDbConnection(true);
}

public static bool EstablishDbConnection(DbConnector connector)
{
   return EstablishDbConnection(null,true);
}

//--------------------------------------------------------------------------------------------------

public static bool EstablishDbConnection(bool deactivateAppWaitingOnError)
{
   return EstablishDbConnection(null,deactivateAppWaitingOnError);
}

public static bool EstablishDbConnection(DbConnector connector,bool deactivateAppWaitingOnError)
{
   connector = connector ?? _DbConnector_;

   using (new XAppWaiting(
      "Подключение к базе данных: [{0}].[{1}] ({2})",
      connector.ServerName, connector.DatabaseName, connector.UserLogName ))
   {
      try
      {
         connector.Connect();
      }
      catch (Exception e)
      {
         return XAppWaiting.ReportError(e,deactivateAppWaitingOnError);
      }
   }

   _DbConnector_ = connector;

   XDispatcher.Synchronize(delegate{
      MainWindow.SetConnectionInformation(); });

   return true;
}

//==================================================================================================

public static bool CapitulateDbConnection(string errorMessage)
{
   XAppWaiting.ReportError
   (
      string.Format(
         "{0}\r\n\r\n"+
         "*** Соединение с SQL будет разорвано! ***",
         errorMessage ),
      false
   );
   XAppWaiting.Activate( AppCommands.DisconnectDatabase.Text );
   DisconnectDatabase();
   Navigate(typeof(StartPage));
   return false;
}

public static bool CapitulateDbConnection( string errorMessageFmt,
                                           params object[] parameters )
{
   return CapitulateDbConnection(string.Format(errorMessageFmt,parameters));
}

public static bool CapitulateDbConnection(Exception e)
{
   return CapitulateDbConnection(XException.GetErrorMessage(e));
}

//==================================================================================================

static App()
{
   Thread.CurrentThread.CurrentCulture =
   Thread.CurrentThread.CurrentUICulture =
      new CultureInfo(XSpecificCultureNames.ru_RU,false);

   XProgram.SpecifyMainThread();
   XProgram.SpecifyInteractsWithDesktop( true );

   XFatalError.Flags = XFatalErrorFlags.ProcessOwnerDefaults;

   XDebug.PrintOperation("Запуск приложения");

   AppDomain.CurrentDomain.ProcessExit += AppDomain_ProcessExit;
}

//--------------------------------------------------------------------------------------------------

static void VerifyPageType(Type pageType)
{
   XArgCheck.VerifyNotNull(pageType);
   if (!XType.IsTypeOf(pageType,typeof(CustomPage)))
      throw new ArgumentException(
         "Ожидался тип страницы \"CustomPage\"." );
}

static void PrepareToNavigation()
{
   XCursor.SetWait();
   XBackgroundOperation.DetachAppWaitingSuppressBackOff();
   XAppWaiting.Activate( null );
}

//--------------------------------------------------------------------------------------------------

static void DoNavigate(CustomPage page)
{
   PrepareToNavigation();
   RoutedUICommand command = null;
   if (page is ConfigurationPage)
      command = AppCommands.OpenConfigurationPage;
   DoNavigate(page,command!=null ? command.Text : null);
}

static void DoNavigate(CustomPage page,string title)
{
   MainWindow.ResetMainMenuContextItem();
   MainWindow.SetPageTitle(
      page.Title =
         title !=null ?
         string.Format("\u00AB {0} \u00BB",title) :
         "" );
   MainWindow.Navigation.Navigate(
      _SelectedPage_ = page );
}

//--------------------------------------------------------------------------------------------------

void MergeDictionary(string relativePath)
{
   Resources.MergedDictionaries.Add(
      (ResourceDictionary)LoadComponent(
         new Uri( relativePath, UriKind.Relative ) ));
}

void MergeDictionaries()
{
   Resources.MergedDictionaries.Clear();

   MergeDictionary("MainCode/Resources/SharedValues.xaml");
   MergeDictionary("MainCode/Resources/CommonResources.xaml");

   MergeDictionary("MainCode/Styles/WindowStyle.xaml");
   MergeDictionary("MainCode/Styles/PageStyle.xaml");
   MergeDictionary(Settings.UseFluctuationEffect ?
                    "MainCode/Styles/FluctuationStyle.xaml" :
                    "MainCode/Styles/FluctuationStyle.Inactive.xaml");
   MergeDictionary("MainCode/Styles/ControlStyles.xaml");
   MergeDictionary("MainCode/Styles/BusyStyles.xaml");
}

void ComplementResources()
{
   var mb = new MultiBinding();
   mb.Bindings.Add( new Binding(){ Path = new PropertyPath("IsCountersRelated") } );
   mb.Converter = new CorrectionCauseToImageConverter();
   Resources.Add( "CorrectionCauseToImageBinding", mb );
}

//--------------------------------------------------------------------------------------------------

void SynchronizeSettings()
{
   XBeeper.IsMute = !Settings.UseBeeperAccompaniment;

   MergeDictionaries();
}

//--------------------------------------------------------------------------------------------------

void CloseCommand_Executed(object sender,ExecutedRoutedEventArgs e)
{
   MainWindow.Close();
}

static void AppDomain_ProcessExit(object sender,EventArgs e)
{
   XDebug.PrintOperation("Завершение процесса");
}

//==================================================================================================

private void Application_Startup(object sender,StartupEventArgs e)
{
   using (new XDebug(MethodBase.GetCurrentMethod()))
   {
      XApplication.SpecifyDispatcher();
      XApplication.AnchorWindow = App.MainWindow;

      XMessageBox.CaptionFormatString = "Мастер: {0}";
      XMessageBox.BeforeShowHandler += delegate
      {
         if ( MainWindow!=null &&
              MainWindow.WindowState==WindowState.Minimized )
            MainWindow.Restore();
      };

      _DbConnector_ = new DbConnector();

      try
      {
         _Settings_ = new AppSettings();
      }
      catch (Exception x)
      {
         XMessageBox.ShowError(
            XException.ComposeChainedMessage( x ),
            "Чтение настроек приложения (пользовательский профиль)" );
         Shutdown( FailureExitCode );
         return;
      }

      SynchronizeSettings();
      ComplementResources();

      XApplication.IsBusyChanged +=
         delegate
         {
            if (!XApplication.IsBusy)
               XCursor.Reset();
         };

      XAppWaiting.DisplayAppWaitingStatusHandler +=
         delegate (bool isActive,string caption,object data)
         {
            MainWindow.SetStatusCaption(caption);
         };

      _StartupArguments_ = e.Args;

      new MainWindow().Show();
   }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // App
//**************************************************************************************************

} // PaycorMaster
