﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.Text;
using System.Globalization;
using System.Runtime.InteropServices;

namespace XHelpers {

//**************************************************************************************************
partial class XConsole {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

enum InputKey { Other, PrintableChar, Backspace, ClearInput, AcceptInput, CancelInput }

//**************************************************************************************************

static string InputRoutine( XConsoleColorRetention colorRetention,
                            [Optional] string initialString,
                            XConsoleInputOptions options,
                            [Optional] XValidateConsoleInputCallback validationCallback )
{
   int
      startPos = Console.CursorLeft,
      maxPos = Console.BufferWidth-1;
   var
      sb = new StringBuilder();

   if (initialString!=null)
   {
      string str =
         StringToPrintable( initialString.Substring( 0,
            Math.Min( initialString.Length, maxPos-startPos ) ) );
      sb.Append(str);
      OutputDefault( (options & XConsoleInputOptions.PasswordMasking)==0 ?
         str : new string(InputPasswordChar,str.Length) );
   }

   ResetInput();

   for (;;)
   {
      InputKey inputKey;
      char printableChar;

      inputKey = ReadInputKey(out printableChar);

      switch (inputKey)
      {
      case InputKey.PrintableChar:
         if (Console.CursorLeft==maxPos)
            goto default;
         OutputDefault(
            (options & XConsoleInputOptions.PasswordMasking)!=0 ?
            InputPasswordChar : printableChar );
         sb.Append(printableChar);
         break;
      case InputKey.Backspace:
         if (Console.CursorLeft==startPos)
            goto default;
         Console.CursorLeft = Console.CursorLeft-1;
         using (new XConsoleColorRetention(null,colorRetention.PreviousBackgroundColor))
            OutputDefault(" ");
         Console.CursorLeft = Console.CursorLeft-1;
         sb.Remove(sb.Length-1,1);
         break;
      case InputKey.ClearInput:
         if (Console.CursorLeft==startPos)
            goto default;
         int savedPos = Console.CursorLeft;
         Console.CursorLeft = startPos;
         using (new XConsoleColorRetention(null,colorRetention.PreviousBackgroundColor))
            OutputDefault(new string(' ',savedPos-startPos));
         Console.CursorLeft = startPos;
         sb.Remove(0,sb.Length);
         break;
      case InputKey.AcceptInput:
         string str=sb.ToString();
         if ( (options & XConsoleInputOptions.RequireNonEmpty)!=0 && sb.Length==0 ||
              validationCallback!=null && !validationCallback(str) )
            goto default;
         return str;
      case InputKey.CancelInput:
         if ((options & XConsoleInputOptions.DisallowCancellation)!=0)
            goto default;
         return null;
      default:
         PlayDisabledSound();
         ResetInput();
         break;
      }
   }
}

//--------------------------------------------------------------------------------------------------

static InputKey ReadInputKey(out char printableChar)
{
   var keyInfo = Console.ReadKey(true);
   var modifiers = keyInfo.Modifiers;

   printableChar = '\0';

   switch (keyInfo.Key)
   {
   case ConsoleKey.Backspace:
      if (modifiers==0)
         return InputKey.Backspace;
      if (modifiers==ConsoleModifiers.Control)
         return InputKey.ClearInput;
      break;
   case ConsoleKey.Enter:
      if (modifiers==0)
         return InputKey.AcceptInput;
      break;
   case ConsoleKey.Escape:
      if (modifiers==0)
         return InputKey.CancelInput;
      break;
   }

   char ch = keyInfo.KeyChar;

   if ( (modifiers==0 || modifiers==ConsoleModifiers.Shift) &&
        IsCharPrintable(ch,false) )
   {
      printableChar = ch;
      return InputKey.PrintableChar;
   }

   return InputKey.Other;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XConsole
//**************************************************************************************************

} // XHelpers
