﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.Threading;
using System.Runtime.InteropServices;

namespace XHelpers {

//**************************************************************************************************
partial class XConsole {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

public static ConsoleKey[] DigitalKeys
{
   get { return XArray.Duplicate(_DigitalKeys_); }
}

public static ConsoleKey[] NumPadKeys
{
   get { return XArray.Duplicate(_NumPadKeys_); }
}

public static XConsoleInputIdlePollingContext InputIdlePollingContext
{
   get
   {
      lock (typeof(XConsole))
         return _InputIdlePollingContext_;
   }
}

//==================================================================================================

/// <exception cref="InvalidOperationException" />
/// <exception cref="XAbortConsoleInputException" />
public static int AcceptInputKey(params ConsoleKey[] keysArray)
{
   return AcceptInputKey(keysArray,null);
}

/// <exception cref="InvalidOperationException" />
/// <exception cref="XAbortConsoleInputException" />
public static int AcceptInputKey( ConsoleKey[] keysArray,
                                  [Optional] ConsoleModifiers[] modifiersArray )
{
   return AcceptInputKey(keysArray,modifiersArray,true);
}

/// <exception cref="InvalidOperationException" />
/// <exception cref="XAbortConsoleInputException" />
public static int AcceptInputKey( ConsoleKey[] keysArray,
                                  [Optional] ConsoleModifiers[] modifiersArray,
                                  bool treatNumPadAsDigital )
{
   XArgCheck.VerifyNotNull(keysArray);

   if ( modifiersArray!=null && keysArray.Length!=modifiersArray.Length ||
        keysArray.Length==0 )
      throw XArgCheck.ArgumentException_IncorrectArguments;

   VerifyIsInVisualMode();

   var indexCapture = new XCapture<int>(-1);

   AcceptInputKey
   (
      (XAcceptConsoleInputKeyCallback)delegate (ConsoleKeyInfo keyInfo)
      {
         for (int i=0; i<keysArray.Length; i++)
         {
            var key = XArray.GetAtZeroBasedIndex(keysArray,i);
            var modifiers = modifiersArray!=null ?
               XArray.GetAtZeroBasedIndex(modifiersArray,i) : 0;

            if ( keyInfo.Modifiers == modifiers &&
                 ( keyInfo.Key == key || IsKeyDigital( key ) &&
                   NumPadKeyAsDigital( keyInfo.Key ) == key ) )
            {
               indexCapture.Value = i;
               return true;
            }
         }
         return false;
      }
   );

   return indexCapture;
}

/// <exception cref="InvalidOperationException" />
/// <exception cref="XAbortConsoleInputException" />
public static ConsoleKeyInfo AcceptInputKey(XAcceptConsoleInputKeyCallback acceptKeyCallback)
{
   lock (typeof(XConsole))
   {
      VerifyIsInVisualMode();
      for (;;)
      {
         ResetInput();
         var keyInfo=Console.ReadKey(true);
         if (acceptKeyCallback(keyInfo))
            return keyInfo;
         PlayDisabledSound();
      }
   }
}

//==================================================================================================

/// <exception cref="InvalidOperationException" />
public static void ResetInput()
{
   lock (typeof(XConsole))
   {
      VerifyIsInVisualMode();
      while (Console.KeyAvailable)
         Console.ReadKey(true);
   }
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="InvalidOperationException" />
/// <exception cref="XAbortConsoleInputException" />
public static void WaitInput()
{
   lock (typeof(XConsole))
      if (InputIdlePollingContext!=null)
      {
         DateTime? prevTime=null;

         while (!Console.KeyAvailable)
         {
            var currTime = DateTime.Now;

            if ( prevTime==null ||
                 XDateTime.ElapsedMilliseconds(prevTime.Value,currTime) >=
                 InputIdlePollingContext.Interval )
            {
               InputIdlePollingContext.Callback();
               prevTime = currTime;
            }

            Thread.Sleep( WaitInputPollingPause );
         }
      }
      else
         while (!Console.KeyAvailable)
            Thread.Sleep( WaitInputPollingPause );
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="InvalidOperationException" />
/// <exception cref="XAbortConsoleInputException" />
public static void AbortConsoleInput()
{
   lock (typeof(XConsole))
   {
      VerifyIsInVisualMode();
      if (InputIdlePollingContext==null)
         throw new InvalidOperationException(
            "Input idle polling context is not active." );
      throw new XAbortConsoleInputException();
   }
}

//==================================================================================================

/// <exception cref="InvalidOperationException" />
/// <exception cref="XAbortConsoleInputException" />
public static ConsoleKeyInfo ReadKey()
{
   lock (typeof(XConsole))
   {
      VerifyIsInVisualMode();
      if (InputIdlePollingContext!=null)
         WaitInput();
      return Console.ReadKey( true );
   }
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="InvalidOperationException" />
public static bool ReadKeyIfAvailable(ref ConsoleKeyInfo keyInfo)
{
   lock (typeof(XConsole))
   {
      VerifyIsInVisualMode();
      if (!Console.KeyAvailable)
         return false;
      keyInfo = Console.ReadKey( true );
      return true;
   }
}

//==================================================================================================

public static bool IsKeyDigital(ConsoleKey key)
{
   return XArray.Contains(_DigitalKeys_,key);
}

public static bool IsKeyNumPad(ConsoleKey key)
{
   return XArray.Contains(_NumPadKeys_,key);
}

//--------------------------------------------------------------------------------------------------

public static ConsoleKey NumPadKeyAsDigital(ConsoleKey key)
{
   return IsKeyNumPad( key ) ?
      (ConsoleKey)( (int)ConsoleKey.D0 + (int)key - (int)ConsoleKey.NumPad0 ) :
      key;
}

//--------------------------------------------------------------------------------------------------

public static string KeyToString(ConsoleKey key)
{
   return KeyToString(key,0);
}

public static string KeyToString(ConsoleKey key,[Optional] ConsoleModifiers modifiers)
{
   if (KeyToStringCallback!=null)
      return KeyToStringCallback(key,modifiers);

   string keyString=key.ToString();

   if ( IsKeyDigital(key) || IsKeyNumPad(key) )
      keyString=keyString[ keyString.Length-1 ].ToString();

   if ((modifiers & ConsoleModifiers.Shift)!=0)
      keyString="Shift+"+keyString;
   if ((modifiers & ConsoleModifiers.Alt)!=0)
      keyString="Alt+"+keyString;
   if ((modifiers & ConsoleModifiers.Control)!=0)
      keyString="Ctrl+"+keyString;

   return string.Format("<{0}>",keyString);
}

//==================================================================================================

public static void PlayDisabledSound()
{
   XSound.PlayDisabled( true );
}

//==================================================================================================

internal static void SetInputIdlePollingContext(XConsoleInputIdlePollingContext context)
{
   lock (typeof(XConsole))
      _InputIdlePollingContext_ = context;
}

//==================================================================================================

static readonly ConsoleKey[]
   _DigitalKeys_=
   {
      ConsoleKey.D0,
      ConsoleKey.D1, ConsoleKey.D2, ConsoleKey.D3,
      ConsoleKey.D4, ConsoleKey.D5, ConsoleKey.D6,
      ConsoleKey.D7, ConsoleKey.D8, ConsoleKey.D9
   },
   _NumPadKeys_=
   {
      ConsoleKey.NumPad0,
      ConsoleKey.NumPad1, ConsoleKey.NumPad2, ConsoleKey.NumPad3,
      ConsoleKey.NumPad4, ConsoleKey.NumPad5, ConsoleKey.NumPad6,
      ConsoleKey.NumPad7, ConsoleKey.NumPad8, ConsoleKey.NumPad9
   };

static XConsoleInputIdlePollingContext
   _InputIdlePollingContext_;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XConsole
//**************************************************************************************************

} // XHelpers
