﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.IO;
using System.Text;
using System.Threading;
using System.Reflection;
using System.Globalization;
using System.Diagnostics;
using System.Runtime.CompilerServices;
using System.ComponentModel;

namespace XHelpers {

//**************************************************************************************************
partial class XFatalError {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

[MethodImpl(MethodImplOptions.NoInlining)]
static string GenerateReport( Assembly assembly, 
                              string cause, string message, Exception e,
                              StackTrace stackTrace )
{
   string
      errorModuleName = XAssembly.GetModuleName(assembly),
      tempFolderPath = Path.GetTempPath();
   if (tempFolderPath==null)
      throw new XOperationFailedException(
         "Can not obtain path to temporary files folder." );
   string
      filePath = Path.Combine
      (
         tempFolderPath,
         string.Format(
            "{0} - FatalErrorReport {1}.htm",
            _AppSimpleName_,
            ErrorTime.ToString("yyyy-MM-dd HH-mm-ss",CultureInfo.InvariantCulture)
         )
      );

   using (TextWriter writer=new StreamWriter(filePath,false,new UTF8Encoding(true)))
   {
      writer.WriteLine("<html>");

      writer.WriteLine("<head>");
      writer.WriteLine("<meta http-equiv=content-type content=\"text/html; charset=utf-8\">");
      writer.WriteLine("<title>{0}</title>",Path.GetFileName(filePath));
      writer.WriteLine("<style>");
      writer.WriteLine("h4{ font-family: \"Trebuchet MS\"; }");
      writer.WriteLine("</style>");
      writer.WriteLine("</head>");

      writer.WriteLine("<body bgcolor=\"Linen\"><font face=\"Trebuchet MS\" size=2>");

      writer.WriteLine(
         "<p><h4 align=\"Center\"><u>Fatal error in <i>&laquo;{0}&raquo;</i> ({1}):</u></h4></p>",
         _MainModuleName_, ErrorTime.ToString("f") );

      if (cause!=null)
         writer.WriteLine(
            "<p><h4 align=\"Center\"><font color=\"Navy\"><i>{0}</i></font></h4></p>",
            cause );

      writer.WriteLine("<p><h4 align=\"Center\">");
      if (!XAppDomain.IsDefault())
      {
         writer.WriteLine(
            "( Application domain: <i>{0}</i> )",
            AppDomain.CurrentDomain.FriendlyName );
         writer.WriteLine("<br>");
      }
      writer.WriteLine(
         "( Source module: <i>{0}</i> )",
         errorModuleName );
      writer.WriteLine("<br>");
      writer.WriteLine(
         "( {0}: <i>ID={1}</i> )",
         XProgram.IsMainThread(_RaiseThread_) ?
            "In main program thread" :
            XProgram.IsMainThreadSpecified ?
               "In secondary program thread" :
               "In following program thread",
         _RaiseThread_.ManagedThreadId );
      writer.WriteLine("</h4></p>");

      writer.WriteLine("<hr><p>");
      if (e!=null)
      {
         if (e.InnerException!=null)
            writer.WriteLine("<ul>");
         for (Exception x=e; x!=null; x=x.InnerException)
         {
            if (x!=e)
               WriteLineToHtm(writer);
            if (e.InnerException!=null)
               writer.Write("<li>");
            WriteLineToHtm( writer,
               "<font color=\"Red\"><b><u>{0}</u></b></font>",
               XException.ComposeMessage(x) );
            WriteLinesToHtm(writer,x.StackTrace);
         }
         if (e.InnerException!=null)
            writer.WriteLine("</ul>");
      }
      else
      {
         writer.WriteLine("<h4><font color=\"Red\"><b>{0}</b></font></h4>",message);
         WriteLinesToHtm(writer,XStackTrace.GetText(stackTrace));
      }
      writer.WriteLine("</p><hr>");

      writer.WriteLine(
         "<p><h4 align=\"Center\"><font color=\"DimGray\" size=3>{0}</font></h4></p>",
         "<i>&laquo; End of fatal error report &raquo;</i>" );

      writer.WriteLine("</body>");

      writer.WriteLine("</html>");
   }

   return filePath;
}

//--------------------------------------------------------------------------------------------------

[MethodImpl(MethodImplOptions.NoInlining)]
static void ShowReport(string filePath)
{
   var startupInfo = new ProcessStartInfo("iexplore.exe",filePath);
   startupInfo.WindowStyle = ProcessWindowStyle.Maximized;

   Process process;

   try
   {
      process = Process.Start( startupInfo );
   }
   catch
   {
      try
      {
         startupInfo = new ProcessStartInfo( filePath );
         process = Process.Start( startupInfo );
      }
      catch (Win32Exception e)
      {
         throw XWin32Error.CreateException(e);
      }
   }

   if (process==null)
      throw new XOperationFailedException(string.Format(
         "Can not open fatal error report (HTML-document).\r\n"+
         "File: \"{0}\".", filePath ));

   process.WaitForInputIdle( 10000 );
}

//==================================================================================================

static void WriteLineToHtm(TextWriter writer)
{
   WriteLineToHtm(writer,"");
}

static void WriteLineToHtm(TextWriter writer,string line)
{
   writer.WriteLine("{0}<br>",line);
}

static void WriteLineToHtm(TextWriter writer,string lineFmt,params object[] parameters)
{
   WriteLineToHtm(writer,string.Format(lineFmt,parameters));
}

//--------------------------------------------------------------------------------------------------

static void WriteLinesToHtm(TextWriter writer,string lines)
{
   if (lines=="")
   {
      WriteLineToHtm(writer);
      return;
   }
   int from=0,to;
   while ((to=lines.IndexOf(Environment.NewLine,from))>=0)
   {
      WriteLineToHtm(writer,lines.Substring(from,to-from));
      from=to+2;
   }
   if (from<lines.Length)
      WriteLineToHtm(writer,lines.Substring(from,lines.Length-from));
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XFatalError
//**************************************************************************************************

} // XHelpers
