﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.IO;
using System.Text;
using System.Threading;
using System.Reflection;
using System.Diagnostics;
using System.Runtime.CompilerServices;

#if SILVERLIGHT
using System.Windows;
using System.Windows.Threading;
#endif

namespace XHelpers {

//**************************************************************************************************
/// <summary>
/// </summary>
public static partial class XFatalError {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class FatalErrorParameters
{
   public
      Assembly Assembly;
   public string
      Cause, Message;
   public Exception
      Exception;
   public StackTrace
      StackTrace;

   internal FatalErrorParameters( Assembly assembly,
                                  string cause, string message,
                                  Exception e, StackTrace stackTrace )
   {
      Assembly=assembly; Cause=cause; Message=message; Exception=e; StackTrace=stackTrace;
   }
}

//**************************************************************************************************

internal static void Access()
{}

//==================================================================================================

static XFatalError()
{
   Library.Access();
}

//--------------------------------------------------------------------------------------------------

[MethodImpl(MethodImplOptions.NoInlining)]
static void RaiseEx(Assembly assembly,string cause,string message,Exception e,StackTrace stackTrace)
{
   Monitor.Enter(typeof(XFatalError));

   if (_IsInTerminationThread_)
      if (_IsInEmergencyHandler_)
         throw new Exception();
      else
         FailOrHang(
            "Reentrance in \"XFatalError.Raise\" method from fatal termination thread." );

   _RaiseThread_ = Thread.CurrentThread;

   lock (_ErrorTimeAccessSyncObject_)
      _ErrorTime_ = DateTime.Now;

#if SILVERLIGHT
   _Dispatcher_ = XApplication.GetDispatcherSpecification();
#endif // SILVERLIGHT

   _ShowReport_ =
      (_Flags_ & XFatalErrorFlags.ShowReport) != 0;

#if !SILVERLIGHT
   if (_ShowReport_)
   {
      bool? interactsWithDesktopSpecification =
         XProgram.GetInteractsWithDesktopSpecification();
      _ShowPreliminaryReport_ =
         interactsWithDesktopSpecification.HasValue &&
         interactsWithDesktopSpecification.Value;
   }
   _TerminateEntireProcess_ =
      (_Flags_ & XFatalErrorFlags.TerminateEntireProcess) != 0 &&
      XSecurityPermission.CheckUnrestricted();
   if (!_TerminateEntireProcess_)
      _ShowReport_ = false;
   _RaiseOnUnhandledException_ =
      (_Flags_ & XFatalErrorFlags.RaiseOnUnhandledException) != 0;
   _UseFatalErrorException_ =
      (_Flags_ & XFatalErrorFlags.UseFatalErrorException) != 0;
#endif // !SILVERLIGHT

   ObtainApplicationName();

   try
   {
#if !SILVERLIGHT
      if (_ShowPreliminaryReport_)
         SetWaitCursor();
#endif // !SILVERLIGHT
      if ( message==null && e==null ||
           message!=null && e!=null ||
           assembly==null )
      {
         cause   = "Internal program error";
         message = "Incorrect call of \"XFatalError.Raise\" method.";
         e = null;
      }
      if (e!=null)
         message=XException.ComposeChainedMessage(e);

      var
         thread = new Thread( TerminationThreadRoutine );
      object
         threadStartParameter = new FatalErrorParameters(assembly,cause,message,e,stackTrace);

#if SILVERLIGHT
      if (_ShowReport_ && _Dispatcher_!=null)
      {
         string
            errorMessage = string.Format
            (
               "___ Fatal error in \u00AB{0}\u00BB: ___\r\n"+
               "\n"+
               "{1}\n"+
               "\n"+
               "*** Application will be terminated! ***",
               _MainModuleName_,
               CombineCauseSourceAndMessage(cause,assembly,message)
            ),
            caption = FatalErrorCaption;
         lock (_ShowReportSyncObject_)
         {
            thread.Start( threadStartParameter );
            Monitor.Wait( _ShowReportSyncObject_ );
            try
            {
               if (XProgram.IsInMainThread())
                  MessageBox.Show(errorMessage,caption,MessageBoxButton.OK);
               else
               {
                  var showErrorCompletedEvent = new ManualResetEvent(false);
                  _Dispatcher_.BeginInvoke((Action)delegate{
                     MessageBox.Show(errorMessage,caption,MessageBoxButton.OK);
                     showErrorCompletedEvent.Set();
                  });
                  showErrorCompletedEvent.WaitOne();
               }
            }
            catch
            {}
            Monitor.Pulse(_ShowReportSyncObject_);
         }
      }
      else
         thread.Start( threadStartParameter );
#else
      thread.Start( threadStartParameter );
      if ( _TerminateEntireProcess_ )
         XThread.HangExecution();
#endif
      thread.Join();
   }
   catch (Exception x)
   {
      FailOrHang(x);
   }

   if (_ReportText_==null)
      FailOrHang("Unknown error in fatal error termination thread.");
   throw RecreateFatalErrorException(
      new XFatalErrorException( _ReportText_ ) );
}

//--------------------------------------------------------------------------------------------------

static void TerminationThreadRoutine(object parametersObject)
{
   _IsInTerminationThread_ = true;
   try
   {
      var errorParameters =
         (FatalErrorParameters)parametersObject;
      PerformFatalError( errorParameters.Assembly,
         errorParameters.Cause, errorParameters.Message,
         errorParameters.Exception, errorParameters.StackTrace );
   }
   catch (XFatalErrorException e)
   {
#if !SILVERLIGHT
      if (_RaiseOnUnhandledException_)
         try
            { DetachUnhandledExceptionHandler(); }
         catch (Exception x)
            { FailOrHang(x); }
#endif // !SILVERLIGHT
      throw RecreateFatalErrorException(e);
   }
   catch (Exception e)
   {
      FailOrHang(e);
   }
}

//--------------------------------------------------------------------------------------------------

static void FailOrHang(string errorText)
{
   if (Interlocked.Exchange(ref _FailOrHangCallFlag_,1)!=0 || _IsInFailOrHang_)
      Thread.Sleep(Timeout.Infinite);

   _IsInFailOrHang_ = true;

#if !SILVERLIGHT
   if (_TerminateEntireProcess_)
      try { FailFast(errorText); }
      catch {}
#endif

   try { Trace(errorText); }
   catch {}

   XThread.HangExecution();
}

static void FailOrHang(Exception e)
{
   FailOrHang(e.ToString());
}

//--------------------------------------------------------------------------------------------------

static Exception RecreateFatalErrorException(XFatalErrorException e)
{
   string errorText =
      Environment.NewLine+e.Message;
#if SILVERLIGHT
   Exception x =
      new XFatalErrorException( errorText );
#else
   Exception x = _UseFatalErrorException_ ?
      new XFatalErrorException( errorText ) :
      new Exception( errorText );
#endif
   x.Data[_FatalErrorExceptionKey_] = true;
   return x;
}

//--------------------------------------------------------------------------------------------------

static void PerformFatalError( Assembly assembly, string cause, string message,
                               Exception e, StackTrace stackTrace )
{
   _ReportText_ = string.Format
   (
      "~~~ Fatal error in \u00AB{0}\u00BB ({1}): ~~~\r\n"+
      "\r\n"+
      "{2}\r\n"+
      "\r\n"+
      "~~~ End of fatal error report. ~~~",
      _MainModuleName_,
      ErrorTime.ToString("f"),
      CombineCauseSourceAndMessage
      (
         cause,
         assembly,
         e != null ?
            XException.ComposeChainedMessage(
               e, true ) :
            string.Format
            (
               "{0}\r\n"+
               "\r\n"+
               "{1}",
               message,
               XStackTrace.GetText( stackTrace )
            )
      )
   );

   OutputTraceMessage( _ReportText_ );

#if !SILVERLIGHT
   string
      reportFilePath = null;
   Exception
      genReportException = null;
   try
   {
      reportFilePath = GenerateReport(assembly,cause,message,e,stackTrace);
   }
   catch (Exception x)
   {
      genReportException = x;
   }
   string uiErrorText = null;
   if (_ShowReport_)
   {
      uiErrorText = string.Format
      (
         "___ Fatal error in \u00AB{0}\u00BB: ___\r\n"+
         "\r\n"+
         "{1}\r\n"+
         "\r\n"+
         "*** The process will be terminated! ***\r\n"+
         "\r\n"+
         "Do you wish to view detailed error report?",
         _MainModuleName_,
         CombineCauseSourceAndMessage(cause,assembly,message)
      );
   }
   TryReportToConsole( cause,
      e!=null ? XException.ComposeChainedMessage(e) : message );
#endif // !SILVERLIGHT

   Thread.MemoryBarrier();

   ProcessEmergencyEvent(_EmergencyHandler_,_ReportText_);

#if SILVERLIGHT
   if (_ShowReport_)
      lock (_ShowReportSyncObject_)
      {
         Monitor.Pulse( _ShowReportSyncObject_ );
         Monitor.Wait( _ShowReportSyncObject_ );
      }

   ProcessEmergencyEvent(_EmergencyFinalHandler_,_ReportText_);

   throw new XFatalErrorException(_ReportText_);
#else
   if (!_TerminateEntireProcess_)
      throw new XFatalErrorException( _ReportText_ );
   if (_ShowReport_)
      if (_ShowPreliminaryReport_)
      {
         Beep();
         if (ShowErrorBox(uiErrorText,FatalErrorCaption,true))
            if (reportFilePath!=null)
               try
               {
                  ShowReport( reportFilePath );
               }
               catch (Exception x)
               {
                  ShowErrorBox(
                     string.Format(
                        "___ Fatal Error Report Showing Error: ___\r\n\r\n"+
                        "{0}", x.Message ),
                     _AppSimpleName_ );
               }
            else
               ShowErrorBox(
                  string.Format(
                     "___ Fatal Error Report Generation Error: ___\r\n\r\n"+
                     "{0}", genReportException.Message ),
                  _AppSimpleName_ );
      }
      else
         if (reportFilePath!=null)
            try
               { ShowReport( reportFilePath ); }
            catch
               {}
   ProcessEmergencyEvent(_EmergencyFinalHandler_,_ReportText_);
   FailFast( _ReportText_ );
#endif
}

//--------------------------------------------------------------------------------------------------

static void ProcessEmergencyEvent(XFatalErrorEmergencyHandler handler,string reportText)
{
   if (handler!=null)
   {
      _IsInEmergencyHandler_ = true;
      try
         { handler( reportText ); }
      catch
         {}
      finally
         { _IsInEmergencyHandler_ = false; }
   }
}

//--------------------------------------------------------------------------------------------------

static string CombineCauseSourceAndMessage(string cause,Assembly assembly,string message)
{
   var sb = new StringBuilder();

   if (cause!=null)
      XTextBuilder.AppendLine(sb,cause);

   XTextBuilder.AppendLine(sb);
#if !SILVERLIGHT
   if (!XAppDomain.IsDefault())
      XTextBuilder.AppendLine( sb,
         "( Application domain: {0} )",
         AppDomain.CurrentDomain.FriendlyName );
   XTextBuilder.AppendLine( sb,
      "( Source module: {0} )",
      XAssembly.GetModuleName(assembly) );
#endif // !SILVERLIGHT
   XTextBuilder.AppendLine( sb,
      "( {0}: ID={1} )",
      XProgram.IsMainThread(_RaiseThread_) ?
         "In the main program thread" :
         XProgram.IsMainThreadSpecified ?
            "In secondary program thread" :
            "In the following program thread",
      _RaiseThread_.ManagedThreadId );

   XTextBuilder.AppendLine(sb,true,message);

   return sb.ToString();
}

//==================================================================================================

static void AppDomain_UnhandledExceptionHandler(object sender,UnhandledExceptionEventArgs e)
{
   ProcessUnhandledException( (Exception)e.ExceptionObject );
}

//==================================================================================================

static XFatalErrorFlags
   _Flags_ = XFatalErrorFlags.Default;

#if SILVERLIGHT
static volatile Dispatcher
   _Dispatcher_;
#endif // SILVERLIGHT

static volatile bool
   _ShowReport_;

#if !SILVERLIGHT
static volatile bool
   _ShowPreliminaryReport_;
static volatile bool
   _TerminateEntireProcess_,
   _RaiseOnUnhandledException_,
   _UseFatalErrorException_;
#endif // !SILVERLIGHT

static XFatalErrorEmergencyHandler
   _EmergencyHandler_,
   _EmergencyFinalHandler_;

//--------------------------------------------------------------------------------------------------

static DateTime
   _ErrorTime_;

static volatile string
   _AppSimpleName_,
   _MainModuleName_,
   _ReportText_;

static bool
   _IsInEmergencyHandler_;
static int
   _FailOrHangCallFlag_;

static volatile Thread
   _RaiseThread_;

[ThreadStatic]
static bool
   _IsInFailOrHang_,
   _IsInTerminationThread_;

//--------------------------------------------------------------------------------------------------

static readonly object
   _FatalErrorExceptionKey_ = new object(),
   _ErrorTimeAccessSyncObject_ = new object();

#if SILVERLIGHT
static readonly object
   _ShowReportSyncObject_ = new object();
#endif // SILVERLIGHT

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XFatalError
//**************************************************************************************************

} // XHelpers
