﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.Threading;
using System.ComponentModel;
using System.Runtime.InteropServices;

namespace XHelpers {

//**************************************************************************************************
/// <summary>
/// </summary>
[AttributeUsage(AttributeTargets.Class)]
public sealed class XSyncOnlyDisposableAttribute: Attribute {}
//**************************************************************************************************


//**************************************************************************************************
/// <summary>
/// </summary>
public abstract class XDisposable: IDisposable {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

public bool IsValidDisposable
{
   get { return _IsValid_; }
}

//--------------------------------------------------------------------------------------------------

public void ValidateDisposable()
{
   _IsValid_ = true;
}

public void Dispose()
{
   if (_IsValid_)
      try
      {
         OnDispose( true );
      }
      finally
      {
         _IsValid_ = false;
         GC.SuppressFinalize( this );
      }
}

//==================================================================================================

public static void Dispose(IDisposable instance)
{
   if (instance!=null)
      instance.Dispose();
}

public static void Dispose<T>([Optional] ref T instance)
{
   if (instance==null)
      return;
   if (!(instance is IDisposable))
      throw new ArgumentException(string.Format(
         "Object type is not derived from \"{0}\" interface.",
         typeof(IDisposable) ));
   var disposable = (IDisposable)instance;
   instance = default(T);
   disposable.Dispose();
}

//==================================================================================================

protected XDisposable()
{}

protected abstract void OnDispose(bool synchronous);

protected void VerifyIsValidDisposable()
{
   if (!IsValidDisposable)
      throw XInvOpExceptionCreation.Create(
         "Object of type \"{0}\" is not valid "+
         "(was not initialized or has already been disposed).",
         this.GetType() );
}

//==================================================================================================

static XDisposable()
{
   Library.Access();
}

//==================================================================================================

~XDisposable()
{
   Thread.MemoryBarrier();
   if (XType.IsMarkedWithAttribute(
      GetType(), typeof(XSyncOnlyDisposableAttribute) ))
   {
      if (_IsValid_)
         throw XInvOpExceptionCreation.ObjectMustBeDisposedSynchronously( this );
   }
   else
      OnDispose( false );
}

//==================================================================================================

bool _IsValid_;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XDisposable
//**************************************************************************************************

} // XHelpers
