﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.Text;
using System.Reflection;
using System.Diagnostics;
using System.Collections;
using System.Collections.Generic;

namespace XHelpers {

//**************************************************************************************************
partial class XException {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

public static string GetMessage(Exception e)
{
   XArgCheck.VerifyNotNull( e );
   return e.Message.Trim()!="" ?
      e.Message : e.GetType().ToString();
}

//==================================================================================================

public static string GetErrorMessage(Exception e)
{
   XArgCheck.VerifyNotNull( e );
   return GetErrorMessage( e,
      GetPackingExceptionTypes() );
}

public static string GetErrorMessage(Exception e,params Type[] packingExceptionTypes)
{
   XArgCheck.VerifyNotNull( e );
   VerifyExceptionArgTypes( packingExceptionTypes );

   var list = new List<Exception>();

   for (Exception x=e; x!=null; x=x.InnerException)
      if (!XType.IsValueOneOf(e,packingExceptionTypes))
         list.Add(x);

   if (list.Count==0)
      return GetMessage(e);

   var sb = new StringBuilder();

   for (int i=0; i<list.Count; i++)
   {
      if (i>0)
      {
         sb.AppendLine();
         sb.AppendLine();
      }
      if (list.Count>1)
         sb.Append("\u2022 ");
      sb.Append(GetMessage(list[i]));
   }

   return sb.ToString();
}

//==================================================================================================

public static string ComposeMessage(Exception e)
{
   return ComposeMessage(e,false);
}

public static string ComposeMessage(Exception e,bool withStackTrace)
{
   string baseMessage =
      string.Format("{0}: {1}",e.GetType(),e.Message);
   if (!withStackTrace)
      return baseMessage;
   return baseMessage +
      (e.StackTrace!=null ? Environment.NewLine+e.StackTrace : "");
}

//--------------------------------------------------------------------------------------------------

public static string ComposeChainedMessage(Exception e)
{
   return ComposeChainedMessage(e,false);
}

public static string ComposeChainedMessage(Exception e,bool withStackTrace)
{
   var sb = new StringBuilder();
   for (Exception x=e; x!=null; x=x.InnerException)
   {
      if (x!=e)
      {
         sb.AppendLine();
         if (withStackTrace)
            sb.AppendLine();
      }
      if (e.InnerException!=null)
         sb.Append("\u2022 ");
      sb.Append( ComposeMessage(x,withStackTrace) );
   }
   return sb.ToString();
}

//==================================================================================================

public static bool IsPackingException(Exception e)
{
   XArgCheck.VerifyNotNull( e );
   lock (typeof(XException))
#if SILVERLIGHT
      return _PackingExceptionTypesList_.Contains( e.GetType() );
#else
      return _PackingExceptionTypesTable_[ e.GetType() ] != null;
#endif
}

//--------------------------------------------------------------------------------------------------

public static Exception SkipPackingExceptions(Exception e)
{
   XArgCheck.VerifyNotNull( e );
   return SkipPackingExceptions(
      e, GetPackingExceptionTypes() );
}

public static Exception SkipPackingExceptions(Exception e,params Type[] packingExceptionTypes)
{
   XArgCheck.VerifyNotNull( e );
   VerifyExceptionArgTypes( packingExceptionTypes );
   for (Exception x=e; x!=null; x=x.InnerException)
      if (!XType.IsValueOneOf(e,packingExceptionTypes))
         return x;
   return e;
}

//--------------------------------------------------------------------------------------------------

public static Type[] GetPackingExceptionTypes()
{
   lock (typeof(XException))
   {
#if SILVERLIGHT
      return _PackingExceptionTypesList_.ToArray();
#else
      var types = new Type[_PackingExceptionTypesTable_.Keys.Count];
      int i=0;
      foreach (object typeObject in _PackingExceptionTypesTable_.Keys)
         types[i++] = (Type)typeObject;
      return types;
#endif
   }
}

//--------------------------------------------------------------------------------------------------

public static void RegisterPackingException(Type exceptionType)
{
   VerifyExceptionArgType( exceptionType );
   lock (typeof(XException))
#if SILVERLIGHT
      if (!XCollection.Contains(_PackingExceptionTypesList_,exceptionType))
         _PackingExceptionTypesList_.Add( exceptionType );
#else
      _PackingExceptionTypesTable_[ exceptionType ] =
         _PackingExceptionTypeMarker_;
#endif
}

public static void RegisterPackingExceptions(params Type[] exceptionTypes)
{
   VerifyExceptionArgTypes( exceptionTypes );
   foreach (Type type in exceptionTypes)
      RegisterPackingException( type );
}

//--------------------------------------------------------------------------------------------------

public static void UnregisterPackingException(Type exceptionType)
{
   VerifyExceptionArgType( exceptionType );
   lock (typeof(XException))
#if SILVERLIGHT
      _PackingExceptionTypesList_.Remove( exceptionType );
#else
      _PackingExceptionTypesTable_[ exceptionType ] = null;
#endif
}

public static void UnregisterPackingExceptions(params Type[] exceptionTypes)
{
   VerifyExceptionArgTypes( exceptionTypes );
   foreach (Type type in exceptionTypes)
      UnregisterPackingException( type );
}

//==================================================================================================

static void VerifyExceptionArgType(Type exceptionType)
{
   XArgCheck.VerifyNotNull( exceptionType );
   XArgCheck.VerifyIsTypeOf( exceptionType, typeof(Exception) );
}

static void VerifyExceptionArgTypes(Type[] exceptionTypes)
{
   XArgCheck.VerifyNotNulls( exceptionTypes );
   foreach (Type type in exceptionTypes)
      VerifyExceptionArgType( type );
}

//==================================================================================================

static XException()
{
   RegisterPackingExceptions(
      typeof(TargetInvocationException),
      typeof(TypeInitializationException) );
}

//==================================================================================================

#if SILVERLIGHT
static readonly List<Type>
   _PackingExceptionTypesList_ = new List<Type>();
#else
static readonly Hashtable
   _PackingExceptionTypesTable_ = new Hashtable();
static readonly object
   _PackingExceptionTypeMarker_ = new object();
#endif

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XException
//**************************************************************************************************

} // XHelpers
