﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.IO;
using System.ComponentModel;

namespace XHelpers {

//**************************************************************************************************
partial class XFile {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

internal static void VerifyPath(string path)
{
   XArgCheck.VerifyStringNotEmpty( path );
   XPath.VerifyNotContainsWildcardChars( path );
}

//==================================================================================================

internal static bool CheckExists(string path,bool directory,out Exception e)
{
   try
      { VerifyExists(path,directory); }
   catch (Exception x)
      { e = x; return false; }
   e = null; return true;
}

internal static void VerifyExists(string path,bool directory)
{
   VerifyPath( path );

   var attributes = File.GetAttributes( path );

   bool isDirectory = (attributes & FileAttributes.Directory) != 0;

   if (directory && !isDirectory || !directory && isDirectory)
      throw CreateNotFoundException(path,directory);
}

//==================================================================================================

internal static bool IsReadonly(string path,bool directory)
{
   return (GetAttributes(path,directory) & FileAttributes.ReadOnly) != 0;
}

//--------------------------------------------------------------------------------------------------

internal static void SetReadonlyAttribute(string path,bool directory)
{
   ModifyAttributes(path,directory,FileAttributes.ReadOnly,0);
}

internal static void ResetReadonlyAttribute(string path,bool directory)
{
   ModifyAttributes(path,directory,0,FileAttributes.ReadOnly);
}

//==================================================================================================

internal static FileAttributes GetAttributes(string path,bool directory)
{
   var attributes = File.GetAttributes( path );

   bool isDirectory = (attributes & FileAttributes.Directory) != 0;

   if (directory && !isDirectory || !directory && isDirectory)
      throw CreateNotFoundException(path,directory);

   return attributes;
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="XOperationFailedException" />
internal static void SetAttributes(string path,bool directory,FileAttributes attributes)
{
   VerifyExists(path,directory);
   File.SetAttributes(path,attributes);
   if (File.GetAttributes(path)!=attributes)
      throw CreateCantSetAttributesException(path,directory);
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="XOperationFailedException" />
public static FileAttributes AddAttributes(string path,bool directory,FileAttributes attributes)
{
   return ModifyAttributes(path,directory,attributes,0);
}

/// <exception cref="XOperationFailedException" />
public static FileAttributes RemoveAttributes(string path,bool directory,FileAttributes attributes)
{
   return ModifyAttributes(path,directory,0,attributes);
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="XOperationFailedException" />
public static FileAttributes ModifyAttributes( string path, bool directory,
                                               FileAttributes attributesToAdd,
                                               FileAttributes attributesToRemove )
{
   VerifyExists(path,directory);

   if ((attributesToRemove & attributesToAdd)!=0)
      throw new ArgumentException(
         "File attributes to set intersect with attributes to remove." );

   FileAttributes
      initialAttributes = GetAttributes(path,directory),
      modifiedAttributes= ( initialAttributes | attributesToAdd ) & ~attributesToRemove;

   if (initialAttributes==modifiedAttributes)
      return initialAttributes;

   File.SetAttributes(path,modifiedAttributes);

   if (File.GetAttributes(path)!=modifiedAttributes)
      throw CreateCantSetAttributesException(path,directory);

   return modifiedAttributes;
}

//==================================================================================================

static FileNotFoundException CreateNotFoundException(string path,bool directory)
{
   return
      new FileNotFoundException(string.Format(
         directory ?
            "Folder not found: \"{0}\"." :
            "File not found: \"{0}\".",
         path ));
}

static XOperationFailedException CreateCantSetAttributesException(string path,bool directory)
{
   return
      new XOperationFailedException(string.Format(
         directory ?
            "Can not correctly set some of attributes to folder: \"{0}\"." :
            "Can not correctly set some of attributes to file: \"{0}\".",
         path ));
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XFile
//**************************************************************************************************

} // XHelpers
