﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.Text;
using System.Collections.Generic;
using System.Reflection;

namespace XHelpers {

//**************************************************************************************************
[Flags]
public enum XEnumParsingOptions
{
   None=0, IgnoreCase=1, SoleValue=2
}
//**************************************************************************************************


//**************************************************************************************************
public static partial class XEnum {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/// <exception cref="XParseValueException" />
public static EnumType Parse<EnumType>(string src) where EnumType: struct
{
   return Parse<EnumType>(src,null);
}

/// <exception cref="XParseValueException" />
public static EnumType Parse<EnumType>(string src,XEnumParsingOptions options)
   where EnumType: struct
{
   return Parse<EnumType>(src,null,options);
}

/// <exception cref="XParseValueException" />
public static EnumType Parse<EnumType>(string src,EnumType def)
   where EnumType: struct
{
   return Parse<EnumType>(src,null,def);
}

/// <exception cref="XParseValueException" />
public static EnumType Parse<EnumType>(string src,XEnumParsingOptions options,EnumType def)
   where EnumType: struct
{
   return Parse<EnumType>(src,null,options,def);
}

//==================================================================================================

/// <exception cref="XParseValueException" />
public static EnumType Parse<EnumType>(string src,char? separator) where EnumType: struct
{
   return Parse<EnumType>(src,separator,XEnumParsingOptions.None,null);
}

/// <exception cref="XParseValueException" />
public static EnumType Parse<EnumType>(string src,char? separator,XEnumParsingOptions options)
   where EnumType: struct
{
   return Parse<EnumType>(src,separator,options,null);
}

/// <exception cref="XParseValueException" />
public static EnumType Parse<EnumType>(string src,char? separator,EnumType def)
   where EnumType: struct
{
   return Parse<EnumType>(src,separator,XEnumParsingOptions.None,def);
}

public static EnumType Parse<EnumType>( string src, char? separator,
                                        XEnumParsingOptions options, EnumType? def )
   where EnumType: struct
{
   Type type = typeof(EnumType);

   XGenericTypeCheck.VerifyIsEnum( type );
   XArgCheck.VerifyNotNull( src );

   string[] stringValues=XString.Split(src,separator ?? ',');

   if (stringValues.Length==0)
      if (def!=null)
         return def.Value;
      else
         throw new ArgumentException(
            "Enumeration value string is empty." );

   if (stringValues.Length>1)
   {
      if (!IsFlagsEnumeration(type))
         throw new ArgumentException(
            "Flags attribute is required for multiple enumeration value." );
      if ((options & XEnumParsingOptions.SoleValue)!=0)
         throw new ArgumentException(
            "Sole value was expected for enumeration." );
   }

   var list = new List<string>();
   var sb = new StringBuilder();

   foreach (string stringValue in stringValues)
   {
      try
      {
         Enum.Parse( type, stringValue,
            (options & XEnumParsingOptions.IgnoreCase)!=0 );
      }
      catch (ArgumentException e)
      {
         throw new XParseValueException( e.Message );
      }

      foreach (string str in list)
         if (XString.Equal( str, stringValue,
                            (options & XEnumParsingOptions.IgnoreCase)!=0 ))
            throw new XParseValueException(string.Format(
               "Dulplicate flag for enumeration value.",
               stringValue ));

      if (list.Count!=0)
         sb.Append(',');
      sb.Append(stringValue);

      list.Add(stringValue);
   }

   try
   {
      return (EnumType)Enum.Parse( type, sb.ToString(),
         (options & XEnumParsingOptions.IgnoreCase)!=0 );
   }
   catch (ArgumentException e)
   {
      throw new XOperationFailedException(e);
   }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XEnum
//**************************************************************************************************

} // XHelpers
