﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.Threading;

namespace XHelpers {

//**************************************************************************************************
/// <summary>
/// </summary>
public enum XAsyncOperationState
{
   NotStarted,
   Started,
   CancellationPending,
   Succeeded,
   Failed,
   Cancelled
}
//**************************************************************************************************


//**************************************************************************************************
/// <summary>
/// </summary>
public partial class XAsyncOperation {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

delegate void AsyncOperationDelegate(object callback,object[] parameters);

//**************************************************************************************************

public XAsyncOperationState State
{
   get { lock (this) return _State_; }
}

//--------------------------------------------------------------------------------------------------

public bool WasStarted
{
   get { return State != XAsyncOperationState.NotStarted; }
}

public bool IsActive
{
   get
   {
      return XOps.IsOneOf( State,
         XAsyncOperationState.Started,
         XAsyncOperationState.CancellationPending );
   }
}

public bool IsFinished
{
   get
   {
      return XOps.IsOneOf( State,
         XAsyncOperationState.Succeeded,
         XAsyncOperationState.Failed,
         XAsyncOperationState.Cancelled );
   }
}

public bool IsCancellationPending
{
   get { return State == XAsyncOperationState.CancellationPending; }
}

public bool Succeeded
{
   get { return State == XAsyncOperationState.Succeeded; }
}

public bool Failed
{
   get { return State == XAsyncOperationState.Failed; }
}

public bool Cancelled
{
   get { return State == XAsyncOperationState.Cancelled; }
}

//--------------------------------------------------------------------------------------------------

public Thread Thread
{
   get { lock (this) return _Thread_; }
}

//--------------------------------------------------------------------------------------------------

#if !SILVERLIGHT
public ThreadPriority Priority
{
   get
   {
      lock (this)
         return IsActive ? Thread.Priority : _InitialPriority_;
   }
   set
   {
      lock (this)
         if (IsActive)
            Thread.Priority = value;
         else
            _InitialPriority_ = value;
   }
}
#endif // !SILVERLIGHT

//--------------------------------------------------------------------------------------------------

public Exception Error
{
   get
   {
      lock (this)
      {
         if (!IsFinished)
            throw new InvalidOperationException(
               "Асинхронная операция не завершена." );
         return _Exception_;
      }
   }
}

//--------------------------------------------------------------------------------------------------

Action FinishCallback
{
   get { lock (this) return _FinishCallback_;  }
   set { lock (this) _FinishCallback_ = value; }
}

//==================================================================================================

public XAsyncOperation(Action operation)
{
   Prepare(operation);
}

public XAsyncOperation(XMultiargDelegate operation,params object[] parameters)
{
   Prepare(operation,parameters);
}

//==================================================================================================

public void Start()
{
   lock (this)
   {
      if (IsActive)
         throw new InvalidOperationException(
            "Asyncronous operation is already active." );
      _State_ = XAsyncOperationState.NotStarted;
      _StartedEvent_.Reset();
      _FinishedEvent_.Reset();
      Thread.MemoryBarrier();
#if SILVERLIGHT
      _Thread_ = new Thread( new ParameterizedThreadStart( OperationThreadRoutine ) );
      _Thread_.Start( _ThreadParameter_ );
#else
      _DoOperationCallback_ =
         new AsyncOperationDelegate( DoOperation );
      _DoOperationCallback_.BeginInvoke(
         _Operation_, _Parameters_,
         EndOperation, _DoOperationCallback_ );
#endif
   }
   _StartedEvent_.WaitOne();
}

//--------------------------------------------------------------------------------------------------

public void Cancel()
{
   Cancel(false);
}

public void Cancel(bool wait)
{
   lock (this)
   {
      VerifyWasStarted();
      if (_State_==XAsyncOperationState.Started)
         _State_ = XAsyncOperationState.CancellationPending;
   }
   if (wait) Wait();
}

//--------------------------------------------------------------------------------------------------

public void Wait()
{
   VerifyWasStarted();
   _FinishedEvent_.WaitOne();
}

public void Wait(TimeSpan timeout)
{
   VerifyWasStarted();
   _FinishedEvent_.WaitOne( timeout );
}

public void Wait(int millisecondsTimeout)
{
   VerifyWasStarted();
   _FinishedEvent_.WaitOne( millisecondsTimeout );
}

//==================================================================================================

void VerifyWasStarted()
{
   if (!WasStarted)
      throw new InvalidOperationException(
         "Asyncronous operation was not started." );
}

//--------------------------------------------------------------------------------------------------

void Prepare(object operation,params object[] parameters)
{
#if SILVERLIGHT
   _ThreadParameter_ = new object[]{operation,parameters};
#else
   _Operation_ = operation;
   _Parameters_ = parameters;
#endif
   _StartedEvent_ = new ManualResetEvent(false);
   _FinishedEvent_= new ManualResetEvent(false);
}

//--------------------------------------------------------------------------------------------------

#if SILVERLIGHT
void OperationThreadRoutine(object parameter)
{
   object[]
      args = (object[])parameter;
   DoOperation(
      args[0], (object[])args[1] );
}
#endif

//--------------------------------------------------------------------------------------------------

void DoOperation(object callback,object[] parameters)
{
   lock (this)
   {
#if !SILVERLIGHT
      _Thread_ = Thread.CurrentThread;
      _Thread_.Priority = _InitialPriority_;
#endif
      _State_ = XAsyncOperationState.Started;
      _StartedEvent_.Set();
   }
#if SILVERLIGHT
   try
   {
#endif
   if (callback is XMultiargDelegate)
      ((XMultiargDelegate)callback)((object[])parameters[0]);
   else
      ((Action)callback)();
#if SILVERLIGHT
   }
   catch (Exception e)
   {
      Finish(e); return;
   }
   Finish();
#endif
}

//--------------------------------------------------------------------------------------------------

#if !SILVERLIGHT
void EndOperation(IAsyncResult ar)
{
   try
      { ((AsyncOperationDelegate)ar.AsyncState).EndInvoke(ar); }
   catch (Exception e)
      { Finish(e); return; }
   Finish();
}
#endif

//--------------------------------------------------------------------------------------------------

void Finish()
{
   Finish(null);
}

void Finish(Exception e)
{
   Thread.MemoryBarrier();
   lock (this)
   {
      _State_ = IsCancellationPending ?
         XAsyncOperationState.Cancelled :
         (_Exception_=e)!=null ?
            XAsyncOperationState.Failed :
            XAsyncOperationState.Succeeded;
      if (_FinishCallback_!=null)
         _FinishCallback_();
   }
   _FinishedEvent_.Set();
}

//==================================================================================================

static XAsyncOperation()
{
   Library.Access();
}

//==================================================================================================

XAsyncOperationState
   _State_;
Action
   _FinishCallback_;
Thread
   _Thread_;
volatile ManualResetEvent
   _StartedEvent_,
   _FinishedEvent_;
Exception
   _Exception_;

//--------------------------------------------------------------------------------------------------

#if SILVERLIGHT
object
   _ThreadParameter_;
#endif // SILVERLIGHT

//--------------------------------------------------------------------------------------------------

#if !SILVERLIGHT
object
   _Operation_;
object[]
   _Parameters_;
AsyncOperationDelegate
   _DoOperationCallback_;
ThreadPriority
   _InitialPriority_ = ThreadPriority.Normal;
#endif // !SILVERLIGHT

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XAsyncOperation
//**************************************************************************************************

} // XHelpers
