﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.Globalization;

namespace XHelpers {

//**************************************************************************************************
/// <summary>
/// </summary>
[Flags]
public enum XHexParseOptions
{
   None                 = 0,
   AllowPrefix          = 1 << 0,
   RequirePrefix        = 1 << 1,
   RequireRegularLength = 1 << 2,
   RequireFullLength    = 1 << 3
}
//**************************************************************************************************


//**************************************************************************************************
/// <summary>
/// </summary>
public static partial class XHexParser {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/// <exception cref="XParseValueException" />
public static T Parse<T>(string src) where T: struct
   { return Parse<T>(src,XHexParseOptions.None); }
/// <exception cref="XParseValueException" />
public static T Parse<T>(string src,T def) where T: struct
   { return Parse<T>(src,XHexParseOptions.None,def); }

/// <exception cref="XParseValueException" />
public static T Parse<T>(string src,XHexParseOptions options) where T: struct
   { return ParseEx(src,options,(T?)null); }
/// <exception cref="XParseValueException" />
public static T Parse<T>(string src,XHexParseOptions options,T def) where T: struct
   { return ParseEx(src,options,(T?)def); }

//--------------------------------------------------------------------------------------------------

/// <exception cref="XParseValueException" />
public static bool TryParse<T>(string src,out T dst) where T: struct
   { return TryParse(src,XHexParseOptions.None,out dst); }
/// <exception cref="XParseValueException" />
public static bool TryParse<T>(string src,T def,out T dst) where T: struct
   { return TryParse(src,XHexParseOptions.None,def,out dst); }

/// <exception cref="XParseValueException" />
public static bool TryParse<T>(string src,out T dst,out XParseValueException e) where T: struct
   { return TryParse(src,XHexParseOptions.None,out dst,out e); }
/// <exception cref="XParseValueException" />
public static bool TryParse<T>( string src, T def,
                                out T dst, out XParseValueException e ) where T: struct
   { return TryParse(src,XHexParseOptions.None,def,out dst,out e); }

/// <exception cref="XParseValueException" />
public static bool TryParse<T>(string src,XHexParseOptions options,out T dst) where T: struct
   { XParseValueException e; return TryParse(src,options,out dst,out e); }
/// <exception cref="XParseValueException" />
public static bool TryParse<T>( string src, XHexParseOptions options, T def,
                                out T dst ) where T: struct
   { XParseValueException e; return TryParse(src,options,def,out dst,out e); }

/// <exception cref="XParseValueException" />
public static bool TryParse<T>( string src, XHexParseOptions options,
                                out T dst, out XParseValueException e ) where T: struct
{
   try
      { dst=Parse<T>(src,options);}
   catch (XParseValueException x)
      { dst=default(T); e=x; return false; }
   e=null; return true;
}
/// <exception cref="XParseValueException" />
public static bool TryParse<T>( string src, XHexParseOptions options, T def,
                                out T dst, out XParseValueException e ) where T: struct
{
   try
      { dst=Parse<T>(src,options,def); }
   catch (XParseValueException x)
      { dst=default(T); e=x; return false; }
   e=null; return true;
}

//==================================================================================================

static XHexParser()
{
   Library.Access();
}

//--------------------------------------------------------------------------------------------------

static T ParseEx<T>(string src,XHexParseOptions options,T? def) where T: struct
{
   Type type = typeof(T);

   if (!XType.IsStdUnsignedNumber(type) && type!=typeof(UIntPtr))
      throw new XGenericTypeException(string.Format(
         "Hexadecimal parse is supported only for following types: "+
         "\"{0}\", \"{1}\", \"{2}\", \"{3}\", \"{4}\".",
         typeof(Byte), typeof(UInt16), typeof(UInt32), typeof(UInt64), typeof(UIntPtr) ));

   if (type==typeof(UIntPtr))
      return (T)(object)new UIntPtr
      (
         XPlatform.Is32BitPlatform ?
            ParseEx<UInt32>( src, options, def!=null ?
               (UInt32?)((UIntPtr)(object)def.Value).ToUInt32() : null ) :
            ParseEx<UInt64>( src, options, def!=null ?
               (UInt64?)((UIntPtr)(object)def.Value).ToUInt64() : null )
      );

   XArgCheck.VerifyNotNull( src );

   src = src.Trim();

   if (src.StartsWith("0x",StringComparison.OrdinalIgnoreCase))
   {
      if ((options & (XHexParseOptions.AllowPrefix | XHexParseOptions.RequirePrefix))==0)
         throw new XParseValueException(
            "Hexadecimal prefix is not allowed." );
      src = src.Remove(0,2);
      if (src=="" || src.StartsWith("0x",StringComparison.OrdinalIgnoreCase))
         throw new XParseValueException(
            "Incorrect hexadecimal number." );
   }
   else
      if ((options & XHexParseOptions.RequirePrefix)!=0)
         throw new XParseValueException(
            "Hexadecimal prefix is required." );

   int length = src.Trim().Length;

   if (length==0)
      throw new XParseValueException(
         "Value string is empty." );

   if ( (options & XHexParseOptions.RequireRegularLength)!=0 &&
        !XBinExponential.IsRegular( length ) )
      throw new XParseValueException(
         "Regular string length is required for hexadecimal number." );

   if ( (options & XHexParseOptions.RequireFullLength)!=0 &&
        !( type==typeof(Byte   ) && length==2  ||
           type==typeof(UInt16 ) && length==4  ||
           type==typeof(UInt32 ) && length==8  ||
           type==typeof(UInt64 ) && length==16 ||
           type==typeof(UIntPtr) && length==UIntPtr.Size*2 ) )
      throw new XParseValueException(
         "Full string length is required for hexadecimal number." );

   return XValueParser.ParseEx<T>( src,
      new object[]{ NumberStyles.HexNumber, NumberFormatInfo.InvariantInfo },
      def );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XHexParser
//**************************************************************************************************

} // XHelpers
