﻿using System;
using System.Data.Linq;

namespace XHelpers.Linq {

//**************************************************************************************************
/// <summary>
/// </summary>
public abstract partial class XDatabaseConnector<TDataContext> where TDataContext: DataContext {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

public string ServerName
{
   get { return _ServerName_; }
   set { VerifyIsNotConnected(); _ServerName_ = value; }
}

public string DatabaseName
{
   get { return _DatabaseName_; }
   set { VerifyIsNotConnected(); _DatabaseName_ = value; }
}

public bool IntegratedSecutity
{
   get
   {
      return _IntegratedSecutity_;
   }
   set
   {
      VerifyIsNotConnected();
      if (_IntegratedSecutity_=value)
      {
         _UserLogName_ = _UserLogName_ ?? XEnvironment.DomainUserName;
         _UserPassword_ = null;
      }
   }
}

public string UserLogName
{
   get { return _UserLogName_; }
   set { VerifyIsNotConnected(); _UserLogName_ = value; }
}

public string UserPassword
{
   get { return _UserPassword_; }
   set { VerifyIsNotConnected(); _UserPassword_ = value; }
}

public string ApplicationName
{
   get { return _ApplicationName_; }
   set { VerifyIsNotConnected(); _ApplicationName_ = value; }
}

public bool UseConnectionPool
{
   get { return _UseConnectionPool_; }
   set { VerifyIsNotConnected(); _UseConnectionPool_ = value; }
}

//--------------------------------------------------------------------------------------------------

public bool IsConnected
{
   get { return _Database_!=null; }
}

public bool IsConnectionBlocked
{
   get { return _IsConnectionBlocked_; }
   set { BlockConnection( value ); }
}

public TDataContext Database
{
   get { VerifyConnectionState(); return _Database_; }
}

//==================================================================================================

public void ResetConnectionProperties()
{
   VerifyIsNotConnected();
   _ServerName_ = _DatabaseName_ = _ApplicationName_ =
   _UserLogName_ = _UserPassword_ = null;
   _IntegratedSecutity_ = false;
   ResetCustomProperties();
}

/// <exception cref="InvalidOperationException" />
public virtual void VerifyConnectionProperties()
{
   if (string.IsNullOrEmpty( ServerName ))
      throw XInvOpExceptionCreation.PropertyWasNotSpecified( "ServerName" );
   if (string.IsNullOrEmpty( DatabaseName ))
      throw XInvOpExceptionCreation.PropertyWasNotSpecified( "DatabaseName" );
   if (!IntegratedSecutity)
   {
      if (string.IsNullOrEmpty( UserLogName ))
         throw XInvOpExceptionCreation.PropertyWasNotSpecified( "UserLogName" );
      if (UserPassword==null)
         throw XInvOpExceptionCreation.PropertyWasNotSpecified( "UserPassword" );
   }
}

//==================================================================================================

public void Connect()
{
   Disconnect();
   var db = CreateDataContext(
      ComposeConnectionString() );
   db.Connection.Open();
   try
   {
      OnEstablishConnection( db );
   }
   catch
   {
      try { OnTerminateConnection( db, false ); }
      catch {}
      try { db.Connection.Close(); }
      catch {}
      throw;
   }
   _Database_ = db;
}

//--------------------------------------------------------------------------------------------------

public void Disconnect()
{
   Disconnect( true );
}

public bool Disconnect(bool throwOnError)
{
   if (!IsConnected)
      return true;
   var connection = _Database_.Connection;
   var db = XOps.DefaultExchange(ref _Database_);
   _IsConnectionBlocked_ = false;
   Exception e = null;
   try
      { OnTerminateConnection( db, true ); }
   catch (Exception x)
      { e = x; }
   try
      { connection.Close(); }
   catch (Exception xx)
      { if (e==null) e = xx; }
   ResetCustomProperties();
   if (e!=null && throwOnError)
      throw e;
   return e==null;
}

public bool Disconnect(out Exception e)
{
   e = null;
   try
      { Disconnect(); }
   catch (Exception x)
      { e = x; }
   return e == null;
}

//--------------------------------------------------------------------------------------------------

public void BlockConnection()
{
   BlockConnection( true );
}

public void BlockConnection(bool value)
{
   VerifyConnectionState();
   _IsConnectionBlocked_ = value;
}

public void UnblockConnection()
{
   BlockConnection( false );
}

//==================================================================================================

public void PerformCriticalOperation(Action operation)
{
   PerformCriticalOperation(operation,false);
}

public void PerformCriticalOperation(Action operation,bool blockConnectionOnFailure)
{
   XArgCheck.VerifyNotNull( operation );
   try
   {
      operation();
   }
   catch
   {
      if (blockConnectionOnFailure)
         BlockConnection();
      else
         Disconnect();
      throw;
   }
}

//==================================================================================================

public void VerifyConnectionState()
{
   if (!IsConnected)
      throw new InvalidOperationException(
         "Database connection has not been established." );
   if (IsConnectionBlocked)
      throw new InvalidOperationException(
         "Database connection is blocked." );
}

public void VerifyIsNotConnected()
{
   if (IsConnected)
      throw new InvalidOperationException(
         "Can not perform operation for connected state of database connector." );
}

//==================================================================================================

protected XDatabaseConnector<TDataContext> Duplicate()
{
   VerifyIsNotConnected();
   return (XDatabaseConnector<TDataContext>)MemberwiseClone();
}

//--------------------------------------------------------------------------------------------------

protected virtual TDataContext CreateDataContext(string connectionString)
{
   throw new NotSupportedException();
}

protected virtual void ResetCustomProperties()
{}

protected virtual string ComposeConnectionString()
{
   VerifyConnectionProperties();

   string appName =
      ApplicationName ?? XProgram.SimpleName;

   return IntegratedSecutity ?
      string.Format
      (
         "Data Source={0};Initial Catalog={1};"+
         "Integrated Security=True;"+
         "Application Name={2};Pooling={3}",
         ServerName, DatabaseName,
         appName, UseConnectionPool
      ) :
      string.Format
      (
         "Data Source={0};Initial Catalog={1};"+
         "User ID={2};Password={3};"+
         "Application Name={4};Pooling={5}",
         ServerName, DatabaseName,
         UserLogName, UserPassword,
         appName, UseConnectionPool
      );
}

//--------------------------------------------------------------------------------------------------

protected virtual void OnEstablishConnection(TDataContext database)
{}

protected virtual void OnTerminateConnection(TDataContext database,bool wasEstablishCompleted)
{}

//==================================================================================================

static XDatabaseConnector()
{
   Library.Access();
}

//==================================================================================================

string
   _ServerName_,
   _DatabaseName_,
   _UserLogName_,
   _UserPassword_,
   _ApplicationName_;

bool
   _IntegratedSecutity_,
   _UseConnectionPool_,
   _IsConnectionBlocked_;

TDataContext
   _Database_;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XDatabaseConnector<TDataContext>
//**************************************************************************************************

} // XHelpers.Linq
