﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

namespace XHelpers {

using System;
using System.Threading;

//**************************************************************************************************
partial class XBeeper {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

public const int

   MinFrequency  = 37,
   MaxFrequency  = Int16.MaxValue,

   MinDuration   = 0,
   MaxDuration   = Int32.MaxValue,

   BaseFrequencyDefault = 440,
   BaseDurationDefault  = 100;

//==================================================================================================

public static int BaseFrequency
{
   get
   {
      return _BaseFrequency_;
   }
   set
   {
      VerifyFrequency( value );
      lock (typeof(XBeeper))
         _BaseFrequency_ = value;
   }
}

public static int BaseDuration
{
   get
   {
      return _BaseDuration_;
   }
   set
   {
      VerifyDuration( value );
      lock (typeof(XBeeper))
         _BaseDuration_ = value;
   }
}

//--------------------------------------------------------------------------------------------------

public static bool IsMute
{
   get
   {
      return _IsMute_;
   }
   set
   {
      lock (typeof(XBeeper))
         _IsMute_ = value;
   }
}

//--------------------------------------------------------------------------------------------------

public static bool IsInitialized
{
   get { return _IsInitialized_; }
}

public static bool IsPlaybackActive
{
   get
   {
      if (Monitor.TryEnter(_PlaybackSyncObject_))
      {
         Monitor.Exit(_PlaybackSyncObject_);
         return true;
      }
      return false;
   }
}

//==================================================================================================

public static bool IsValidFrequency(int frequency)
{
   return XOps.IsBetween(frequency,MinFrequency,MaxFrequency);
}

public static bool IsValidDuration(int duration)
{
   return XOps.IsBetween(duration,MinDuration,MaxDuration);
}

//--------------------------------------------------------------------------------------------------

public static void VerifyFrequency(int frequency)
{
   if (!IsValidFrequency( frequency ))
      throw new ArgumentException( string.Format(
         "Incorrect beep frequency: {0} (Hz).\r\n"+
         "Valid range: {1}..{2} (Hz).",
         frequency, MinFrequency, MaxFrequency ));
}

public static void VerifyDuration(int duration)
{
   if (!IsValidFrequency( duration ))
      throw new ArgumentException( string.Format(
         "Incorrect beep duration: {0} (ms).\r\n"+
         "Valid range: {1}..{2}.",
         duration, MinDuration, MaxDuration ));
}

//==================================================================================================

public static void Initialize()
{
   lock (_InitializationSyncObject_)
      if (!_IsInitialized_)
      {
         _PlaybackThread_ = new Thread( PlaybackThread );
         _PlaybackThread_.Priority = _DefaultPriority_;
         _PlaybackThread_.IsBackground = true;

         _StartedEvent_ = new AutoResetEvent(false);

         lock (_ThreadActivitySyncObject_)
         {
            _PlaybackThread_.Start();
            Monitor.Wait(_ThreadActivitySyncObject_);
         }

         _IsInitialized_ = true;
      }
}

//==================================================================================================

public static void Play(params double[] beeps)
{
   Play(false,beeps);
}

public static void Play(bool synchronously,params double[] beeps)
{
   int[] beepsToPlay;

   lock (typeof(XBeeper))
   {
      if (beeps==null || IsMute)
         return;
      beepsToPlay = PrepareBeeps( beeps );
   }

   Monitor.Enter(_StopSignalSyncObject_);

   lock (_StartSyncObject_)
   {
      Initialize();

      lock (_ThreadActivitySyncObject_)
      {
         Monitor.Exit(_StopSignalSyncObject_);

         _PlaybackBeeps_ = beepsToPlay;

         _PlaybackThread_.Priority =
            XOps.IsOneOf( Thread.CurrentThread.Priority,
               ThreadPriority.AboveNormal,
               ThreadPriority.Highest ) ?
            ThreadPriority.Highest : _DefaultPriority_;

         Monitor.Pulse(_ThreadActivitySyncObject_);
      }
      _StartedEvent_.WaitOne();

      if (synchronously)
         WaitActivePlayback();
   }
}

//--------------------------------------------------------------------------------------------------

public static void SyncPlay(params double[] beeps)
{
   Play(true,beeps);
}

//==================================================================================================

public static bool StopActivePlayback()
{
   return StopActivePlayback( true );
}

public static bool StopActivePlayback(bool waitPlaybackFinish)
{
   if (!waitPlaybackFinish)
   {
      _StopFlag_ = true;
      return false;
   }
   lock (_StopControlSyncObject_)
   {
      _StopInitiatorThread_= Thread.CurrentThread;
      _StopOwnerThread_ = null;
      lock ( _StopSignalSyncObject_ )
         WaitActivePlayback();
      return _StopOwnerThread_ == Thread.CurrentThread;
   }
}

//--------------------------------------------------------------------------------------------------

public static void WaitActivePlayback()
{
   lock (_PlaybackSyncObject_) {}
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XBeeper
//**************************************************************************************************

} // XHelpers
