﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.IO;
using System.Xml;
using System.Xml.Schema;
using System.Collections.Generic;

namespace XHelpers {

//**************************************************************************************************
/// <summary>
/// </summary>
public abstract partial class XXmlConfigurationBase {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

protected XmlNode CurrentNode;

//==================================================================================================

protected XXmlConfigurationBase(string schemaText)
{
   using (var stringReader=new StringReader(schemaText))
      _Schema_ = XmlSchema.Read(stringReader,null);
}

//--------------------------------------------------------------------------------------------------

protected XmlDocument Document
{
   get { return _Document_; }
}

//--------------------------------------------------------------------------------------------------

protected XDictionary<string,string> AttributeValues
{
   get
   {
      VerifyCurrentNode();
      return GetAttributeValues(CurrentNode);
   }
}

protected XmlNode[] ChildElements
{
   get
   {
      VerifyCurrentNode();
      return GetChildElements(CurrentNode);
   }
}

//==================================================================================================

protected void LoadDocument(string fileName)
{
   var schemaSet = new XmlSchemaSet();
   var readerSettings = new XmlReaderSettings();

   schemaSet.Add( _Schema_ );

   readerSettings.ValidationType = ValidationType.Schema;
   readerSettings.Schemas = schemaSet;

   _DocumentFilePath_ = fileName;

   try
   {
      _DocumentFilePath_ = Path.GetFullPath(fileName);

      using (XmlReader reader=XmlReader.Create(_DocumentFilePath_,readerSettings))
         _Document_.Load(reader);

      if ( _Document_.DocumentElement.NamespaceURI != _Schema_.TargetNamespace )
         throw new XXmlConfigurationException(string.Format(
            "XML element of the document must belong to following namespace: \"{0}\".",
            _Schema_.TargetNamespace ));
   }
   catch (FileNotFoundException e)
   {
      throw new XXmlConfigurationException( e.Message );
   }
   catch (Exception e)
   {
      throw new XXmlConfigurationException(string.Format(
         "{0}\r\n"+
         "File: \"{1}\".",
         e.Message, _DocumentFilePath_ ));
   }

   _NamespaceManager_ = new XmlNamespaceManager(_Document_.NameTable);
   _NamespaceManager_.AddNamespace("x",_Schema_.TargetNamespace);

   CurrentNode = _Document_.DocumentElement;
}

//--------------------------------------------------------------------------------------------------

protected void NavigateToNode(string docElementRelativePath)
{
   if (!TryNavigateToNode( docElementRelativePath ))
      throw CreateExceptionWithFile(
         "XML node is not found: \"/x:{0}/{1}\".",
         _Document_.DocumentElement.Name,
         docElementRelativePath );
}

protected bool TryNavigateToNode(string docElementRelativePath)
{
   string nodePath =
      string.Format( "/x:{0}/{1}",
         _Document_.DocumentElement.Name,
         docElementRelativePath );

   CurrentNode = _Document_.SelectSingleNode(nodePath,_NamespaceManager_);
   _CurrentNodePath_ = CurrentNode!=null ? nodePath : null;

   return CurrentNode != null;
}

//--------------------------------------------------------------------------------------------------

protected void VerifyCurrentNode()
{
   if (CurrentNode==null)
      throw new InvalidOperationException(
         "Current XML node was not specified." );
}

//==================================================================================================

protected XXmlConfigurationException CreateExceptionWithPath(string message)
{
   VerifyCurrentNode();
   return
      new XXmlConfigurationException(string.Format(
         "{0}\r\n"+
         "File: \"{1}\".\r\n"+
         "Node path: \"{2}\".",
         message, _DocumentFilePath_, _CurrentNodePath_
      ));
}

protected XXmlConfigurationException CreateExceptionWithPath( string messageFmt,
                                                              params object[] parameters )
{
   return CreateExceptionWithPath(string.Format(messageFmt,parameters));
}

//--------------------------------------------------------------------------------------------------

protected XXmlConfigurationException CreateExceptionWithPath(XmlAttribute attribute,string message)
{
   VerifyCurrentNode();
   return
      new XXmlConfigurationException(string.Format(
         "{0}\r\n"+
         "File: \"{1}\".\r\n"+
         "Node path: \"{2}\".\r\n"+
         "Attribute: \"{3}\".",
         message, _DocumentFilePath_, _CurrentNodePath_ , attribute.Name
      ));
}

protected XXmlConfigurationException CreateExceptionWithPath( XmlAttribute attribute,
                                                              string messageFmt,
                                                              params object[] parameters )
{
   return CreateExceptionWithPath(attribute,string.Format(messageFmt,parameters));
}

//==================================================================================================

protected XXmlConfigurationException CreateExceptionWithFile(string message)
{
   return
      new XXmlConfigurationException(string.Format(
         "{0}\r\n"+
         "File: \"{1}\".",
         message, _DocumentFilePath_, _CurrentNodePath_
      ));
}

protected XXmlConfigurationException CreateExceptionWithFile( string messageFmt,
                                                              params object[] parameters )
{
   return CreateExceptionWithFile(string.Format(messageFmt,parameters));
}

//==================================================================================================

protected static XDictionary<string,string> GetAttributeValues(XmlNode node)
{
   var namedStrings = new XDictionary<string,string>();

   foreach (XmlAttribute attribute in node.Attributes)
      namedStrings.Add(attribute.Name,attribute.Value);

   return namedStrings;
}

protected static XmlNode[] GetChildElements(XmlNode node)
{
   var nodesList = new List<XmlNode>();

   foreach (XmlNode childNode in nodesList)
      if (childNode.NodeType==XmlNodeType.Element)
         nodesList.Add(childNode);

   return nodesList.ToArray();
}

//==================================================================================================

static XXmlConfigurationBase()
{
   Library.Access();
}

//==================================================================================================

XmlSchema
   _Schema_;
XmlDocument
   _Document_ = new XmlDocument();
XmlNamespaceManager
   _NamespaceManager_;
string
   _DocumentFilePath_,
   _CurrentNodePath_;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XXmlConfigurationBase
//**************************************************************************************************

} // XHelpers
