﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.ComponentModel;
using System.Runtime.InteropServices;

namespace XHelpers {

//**************************************************************************************************
/// <summary>
/// </summary>
public static partial class XMemory {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

public static unsafe IntPtr Clear([Out] IntPtr destination,int lengthInBytes)
{
   XArgCheck.VerifyNotZeroPointer( destination );
   for (var pby=(byte*)destination; pby<(byte*)((int)destination+lengthInBytes); pby++)
      *pby = 0;
   return destination;
}

public static unsafe IntPtr Copy([In] IntPtr source,[Out] IntPtr destination,int lengthInBytes)
{
   XArgCheck.VerifyNotZeroPointer( source );
   XArgCheck.VerifyNotZeroPointer( destination );
   for (int i=0; i<lengthInBytes; i++)
      *(byte*)((int)destination+i) = *(byte*)((int)source+i);
   return destination;
}

//==================================================================================================

/// <exception cref="XOperationFailedException" />
public static IntPtr Allocate(int sizeInBytes)
{
   VerifyAllocationSize( sizeInBytes );
   IntPtr allocatedBlock = Win32.HeapAlloc(
      XProcess.HeapHandle, Win32.HEAP_ZERO_MEMORY, (UIntPtr)sizeInBytes );
   if (allocatedBlock==IntPtr.Zero)
      throw new XOperationFailedException(
         "Can not allocate memory block." );
   return allocatedBlock;
}

/// <exception cref="XOperationFailedException" />
public static IntPtr ReAllocate(IntPtr allocatedBlock,int newSizeInBytes)
{
   XArgCheck.VerifyNotZeroPointer( allocatedBlock );
   VerifyAllocationSize( newSizeInBytes );
   if (GetAllocatedSize(allocatedBlock)==newSizeInBytes)
      return allocatedBlock;
   IntPtr newBlock = Win32.HeapReAlloc(
      XProcess.HeapHandle, Win32.HEAP_ZERO_MEMORY,
      allocatedBlock, (UIntPtr)newSizeInBytes );
   if (newBlock==IntPtr.Zero)
      throw new XOperationFailedException(
         "Can not allocate memory block." );
   return newBlock;
}

/// <exception cref="XOperationFailedException" />
public static IntPtr ReAllocate(ref IntPtr allocatedBlock,int newSizeInBytes)
{
   return allocatedBlock = ReAllocate(allocatedBlock,newSizeInBytes);
}

/// <exception cref="XOperationFailedException" />
public static IntPtr Duplicate([In] IntPtr allocatedBlock)
{
   int size = GetAllocatedSize( allocatedBlock );
   return Copy(allocatedBlock,Allocate(size),size);
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="XOperationFailedException" />
public static int GetAllocatedSize([In] IntPtr allocatedBlock)
{
   XArgCheck.VerifyNotZeroPointer( allocatedBlock );
   UIntPtr heapSizeResult = Win32.HeapSize(
      XProcess.HeapHandle, 0, allocatedBlock );
   int size = unchecked(
      (int)heapSizeResult );
   if (size==-1)
      throw new XOperationFailedException(
         "Can not determine size of previously allocated memory block." );
   if (size<0)
      throw new XOperationFailedException(
         "Previously allocated memory block is too large (for signed integer value)." );
   return size;
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="Win32Exception" />
public static void Free([Optional] IntPtr allocatedBlock)
{
   if ( allocatedBlock!=IntPtr.Zero &&
        !Win32.HeapFree( XProcess.HeapHandle, 0, allocatedBlock ) )
      throw XWin32Error.CreateExceptionFromLastError();
}

/// <exception cref="Win32Exception" />
public static void Free([Optional] ref IntPtr allocatedBlock)
{
   Free( XIntPtr.InterlockedZeroExchange(ref allocatedBlock) );
}

//==================================================================================================

static XMemory()
{
   Library.Access();
}

//--------------------------------------------------------------------------------------------------

static void VerifyAllocationSize(int sizeInBytes)
{
   if (sizeInBytes<0)
      throw new ArgumentException(
         "Incorrect size of memory allocation block." );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XMemory
//**************************************************************************************************

} // XHelpers
