﻿//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//
// XHelpers Solution Library (Freeware);  Author: Sergei Kitaev
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

using System;
using System.ComponentModel;
using System.Runtime.InteropServices;

namespace XHelpers {

//**************************************************************************************************
/// <summary>
/// </summary>
public static partial class XSZ {
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

public static void VerifyCapacity(int capacity)
{
   if (capacity<1)
      throw new ArgumentOutOfRangeException( null,
         "Incorrect capacity for zero terminated string." );
}

public static unsafe string ToString([Optional][In] IntPtr psz)
{
   return psz!=null ? new string((char*)psz) : null;
}

//==================================================================================================

public static unsafe int Length([In] IntPtr psz)
{
   XArgCheck.VerifyNotZeroPointer( psz );
   for (char* pch=(char*)psz;;pch++)
      if (*pch=='\0')
         return (int)((uint)pch-(uint)psz);
}

public static bool IsEmpty([In] IntPtr psz)
{
   return Length( psz )==0;
}

public static bool IsNullOrEmpty([In] IntPtr psz)
{
   return psz==IntPtr.Zero || IsEmpty(psz);
}

//--------------------------------------------------------------------------------------------------

public static unsafe IntPtr Copy([In] IntPtr pszSource,[Out] IntPtr pszDestination)
{
   XArgCheck.VerifyNotZeroPointer( pszSource );
   XArgCheck.VerifyNotZeroPointer( pszDestination );
   for ( char* pchSource=(char*)pszSource,
               pchDestination=(char*)pszDestination;;
         pchSource++, pchDestination++ )
      if ((*pchDestination=*pchSource)=='\0')
         return pszDestination;
}

/// <exception cref="XOperationFailedException" />
public static IntPtr Copy([In] IntPtr pszSource,[Out] IntPtr pszDestination,int destBuffCapacity)
{
   XArgCheck.VerifyNotZeroPointer( pszSource );
   XArgCheck.VerifyNotZeroPointer( pszDestination );
   if (destBuffCapacity<=1)
      throw new ArgumentException(
         "Wrong destination capacity for zero-terminated string." );
   int sourceLength = Length(pszSource);
   if (sourceLength>=destBuffCapacity)
      throw new XOperationFailedException(
         "Insuffisient buffer for zero-terminated string." );
   return XMemory.Copy<char>(pszSource,pszDestination,sourceLength+1);
}

//==================================================================================================

/// <exception cref="XOperationFailedException" />
public static unsafe IntPtr Allocate(int capacity)
{
   VerifyCapacity( capacity );
   IntPtr result = XMemory.Allocate<char>( capacity );
   ((char*)result)[ capacity-1 ] = '\0';
   return result;
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="XOperationFailedException" />
public static unsafe IntPtr ReAllocate(IntPtr allocatedBuffer,int newCapacity)
{
   VerifyCapacity( newCapacity );
   IntPtr result = XMemory.ReAllocate<char>(allocatedBuffer,newCapacity);
   ((char*)result)[ newCapacity-1 ] = '\0';
   return result;
}

/// <exception cref="XOperationFailedException" />
public static unsafe IntPtr ReAllocate(ref IntPtr allocatedBuffer,int newCapacity)
{
   VerifyCapacity( newCapacity );
   return allocatedBuffer = ReAllocate(allocatedBuffer,newCapacity);
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="XOperationFailedException" />
public static IntPtr ReAllocateByLength(IntPtr pszAllocatedSZ)
{
   return ReAllocate(pszAllocatedSZ,Length(pszAllocatedSZ));
}

/// <exception cref="XOperationFailedException" />
public static IntPtr ReAllocateByLength(ref IntPtr pszAllocatedSZ)
{
   return pszAllocatedSZ = ReAllocateByLength( pszAllocatedSZ );
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="XOperationFailedException" />
public static IntPtr Duplicate([In] IntPtr psz)
{
   return Copy(psz,Allocate(Length(psz)));
}

/// <exception cref="XOperationFailedException" />
public static unsafe IntPtr Duplicate(string str)
{
   XArgCheck.VerifyNotNull( str );
   char* buffer = XUnsafeCharBuffer.Allocate( str.Length+1 );
   for (int i=0; i<str.Length; i++)
      buffer[i] = str[i];
   buffer[ str.Length ] = '\0';
   return (IntPtr)buffer;
}

/// <exception cref="XOperationFailedException" />
public static unsafe IntPtr Duplicate([In] char* psz)
{
   return Duplicate((IntPtr)psz);
}

/// <exception cref="XOperationFailedException" />
public static IntPtr DuplicateAllocatedBuffer([In] IntPtr allocatedBuffer)
{
   return XMemory.Duplicate<char>( allocatedBuffer );
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="XOperationFailedException" />
public static int GetAllocatedCapacity([In] IntPtr allocatedBuffer)
{
   return XMemory.GetAllocatedCapacity<char>( allocatedBuffer );
}

//--------------------------------------------------------------------------------------------------

/// <exception cref="Win32Exception" />
public static void Free([Optional] IntPtr allocatedBuffer)
{
   XMemory.Free<char>( allocatedBuffer );
}

/// <exception cref="Win32Exception" />
public static void Free([Optional] ref IntPtr allocatedBuffer)
{
   XMemory.Free(ref allocatedBuffer);
}

//==================================================================================================

static XSZ()
{
   Library.Access();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
} // XSZ
//**************************************************************************************************

} // XHelpers
